@@ -3,6 +3,9 @@
 #include <algorithm>
 #include <array>
 
+// SP + 8 points to the first argument that is passed on the stack
+#define ARG0_STACK 8
+
 namespace bpftrace {
 namespace arch {
 
@@ -29,12 +32,6 @@ static std::array<std::string, 27> registers = {
   "flags",
   "sp",
   "ss",
-  "fs_base",
-  "gs_base",
-  "ds",
-  "es",
-  "fs",
-  "gs",
 };
 
 static std::array<std::string, 6> arg_registers = {
@@ -80,6 +77,11 @@ int sp_offset()
   return offset("sp");
 }
 
+int arg_stack_offset()
+{
+  return ARG0_STACK / 8;
+}
+
 std::string name()
 {
   return std::string("x86_64");
