@@ -93,6 +93,7 @@ void Identifier::accept(Visitor &v) {
 PositionalParameter::PositionalParameter(PositionalParameterType ptype, long n)
     : ptype(ptype), n(n)
 {
+  is_literal = true;
 }
 
 PositionalParameter::PositionalParameter(PositionalParameterType ptype,
@@ -100,6 +101,7 @@ PositionalParameter::PositionalParameter(PositionalParameterType ptype,
                                          location loc)
     : Expression(loc), ptype(ptype), n(n)
 {
+  is_literal = true;
 }
 
 void PositionalParameter::accept(Visitor &v) {
@@ -184,7 +186,7 @@ Unop::Unop(int op, Expression *expr, location loc)
 {
 }
 
-Unop::Unop(int op, Expression *expr, bool is_post_op, location)
+Unop::Unop(int op, Expression *expr, bool is_post_op, location loc)
     : Expression(loc), expr(expr), op(op), is_post_op(is_post_op)
 {
 }
@@ -222,6 +224,11 @@ FieldAccess::FieldAccess(Expression *expr,
 {
 }
 
+FieldAccess::FieldAccess(Expression *expr, ssize_t index, location loc)
+    : Expression(loc), expr(expr), index(index)
+{
+}
+
 void FieldAccess::accept(Visitor &v) {
   v.visit(*this);
 }
@@ -257,6 +264,16 @@ void Cast::accept(Visitor &v) {
   v.visit(*this);
 }
 
+Tuple::Tuple(ExpressionList *elems, location loc)
+    : Expression(loc), elems(elems)
+{
+}
+
+void Tuple::accept(Visitor &v)
+{
+  v.visit(*this);
+}
+
 Statement::Statement(location loc) : Node(loc)
 {
 }
@@ -314,98 +331,8 @@ void Predicate::accept(Visitor &v) {
   v.visit(*this);
 }
 
-AttachPoint::AttachPoint(const std::string &provider, location loc)
-    : Node(loc), provider(probetypeName(provider))
-{
-}
-
-AttachPoint::AttachPoint(const std::string &provider,
-                         const std::string &func,
-                         location loc)
-    : Node(loc),
-      provider(probetypeName(provider)),
-      func(func),
-      need_expansion(true)
-{
-}
-
-AttachPoint::AttachPoint(const std::string &provider,
-                         const std::string &target,
-                         const std::string &func,
-                         bool need_expansion,
-                         location loc)
-    : Node(loc),
-      provider(probetypeName(provider)),
-      target(target),
-      func(func),
-      need_expansion(need_expansion)
-{
-}
-
-AttachPoint::AttachPoint(const std::string &provider,
-                         const std::string &target,
-                         const std::string &ns,
-                         const std::string &func,
-                         bool need_expansion,
-                         location loc)
-    : Node(loc),
-      provider(probetypeName(provider)),
-      target(target),
-      ns(ns),
-      func(func),
-      need_expansion(need_expansion)
-{
-}
-
-AttachPoint::AttachPoint(const std::string &provider,
-                         const std::string &str,
-                         uint64_t val,
-                         location loc)
-    : Node(loc), provider(probetypeName(provider)), need_expansion(true)
-{
-  if (this->provider == "uprobe" || this->provider == "uretprobe")
-  {
-    target = str;
-    address = val;
-  }
-  else if (this->provider == "kprobe")
-  {
-    func = str;
-    func_offset = val;
-  }
-  else
-  {
-    target = str;
-    freq = val;
-  }
-}
-
-AttachPoint::AttachPoint(const std::string &provider,
-                         const std::string &target,
-                         uint64_t addr,
-                         uint64_t len,
-                         const std::string &mode,
-                         location loc)
-    : Node(loc),
-      provider(probetypeName(provider)),
-      target(target),
-      addr(addr),
-      len(len),
-      mode(mode)
-{
-}
-
-AttachPoint::AttachPoint(const std::string &provider,
-                         const std::string &target,
-                         const std::string &func,
-                         uint64_t offset,
-                         location loc)
-    : Node(loc),
-      provider(probetypeName(provider)),
-      target(target),
-      func(func),
-      need_expansion(true),
-      func_offset(offset)
+AttachPoint::AttachPoint(const std::string &raw_input, location loc)
+    : Node(loc), raw_input(raw_input)
 {
 }
 
@@ -426,7 +353,8 @@ void If::accept(Visitor &v) {
   v.visit(*this);
 }
 
-Unroll::Unroll(long int var, StatementList *stmts) : var(var), stmts(stmts)
+Unroll::Unroll(Expression *expr, StatementList *stmts, location loc)
+    : Statement(loc), expr(expr), stmts(stmts)
 {
 }
 
@@ -441,6 +369,16 @@ Probe::Probe(AttachPointList *attach_points,
 {
 }
 
+void While::accept(Visitor &v)
+{
+  v.visit(*this);
+}
+
+void Jump::accept(Visitor &v)
+{
+  v.visit(*this);
+}
+
 void Probe::accept(Visitor &v) {
   v.visit(*this);
 }
@@ -454,6 +392,21 @@ void Program::accept(Visitor &v) {
   v.visit(*this);
 }
 
+std::string opstr(Jump &jump)
+{
+  switch (jump.ident)
+  {
+    case bpftrace::Parser::token::RETURN:
+      return "return";
+    case bpftrace::Parser::token::BREAK:
+      return "break";
+    case bpftrace::Parser::token::CONTINUE:
+      return "continue";
+    default:
+      throw std::runtime_error("Unknown jump");
+  }
+}
+
 std::string opstr(Binop &binop)
 {
   switch (binop.op) {
@@ -513,6 +466,10 @@ std::string AttachPoint::name(const std::string &attach_point) const
     n += ":" + std::to_string(address);
   if (freq != 0)
     n += ":" + std::to_string(freq);
+  if (len != 0)
+    n += ":" + std::to_string(len);
+  if (mode.size())
+    n += ":" + mode;
   return n;
 }
 
