@@ -11,7 +11,10 @@ namespace ast {
 class FieldAnalyser : public Visitor {
 public:
   explicit FieldAnalyser(Node *root, BPFtrace &bpftrace)
-    : type_(""), root_(root), bpftrace_(bpftrace)
+      : type_(""),
+        root_(root),
+        bpftrace_(bpftrace),
+        prog_type_(BPF_PROG_TYPE_UNSPEC)
   { }
 
   void visit(Integer &integer) override;
@@ -29,11 +32,14 @@ public:
   void visit(FieldAccess &acc) override;
   void visit(ArrayAccess &arr) override;
   void visit(Cast &cast) override;
+  void visit(Tuple &tuple) override;
   void visit(ExprStatement &expr) override;
   void visit(AssignMapStatement &assignment) override;
   void visit(AssignVarStatement &assignment) override;
   void visit(If &if_block) override;
   void visit(Unroll &unroll) override;
+  void visit(While &while_block) override;
+  void visit(Jump &jump) override;
   void visit(Predicate &pred) override;
   void visit(AttachPoint &ap) override;
   void visit(Probe &probe) override;
@@ -45,6 +51,10 @@ private:
   std::string    type_;
   Node          *root_;
   BPFtrace      &bpftrace_;
+  bpf_prog_type  prog_type_;
+  bool           builtin_args_;
+
+  std::map<std::string, SizedType> ap_args_;
 };
 
 } // namespace ast
