@@ -18,8 +18,21 @@ public:
   explicit SemanticAnalyser(Node *root,
                             BPFtrace &bpftrace,
                             BPFfeature &feature,
-                            std::ostream &out = std::cerr)
-      : root_(root), bpftrace_(bpftrace), feature_(feature), out_(out)
+                            std::ostream &out = std::cerr,
+                            bool has_child = true)
+      : root_(root),
+        bpftrace_(bpftrace),
+        feature_(feature),
+        out_(out),
+        has_child_(has_child)
+  {
+  }
+
+  explicit SemanticAnalyser(Node *root,
+                            BPFtrace &bpftrace,
+                            BPFfeature &feature,
+                            bool has_child)
+      : SemanticAnalyser(root, bpftrace, feature, std::cerr, has_child)
   {
   }
 
@@ -34,10 +47,13 @@ public:
   void visit(Variable &var) override;
   void visit(Binop &binop) override;
   void visit(Unop &unop) override;
+  void visit(While &while_block) override;
+  void visit(Jump &jump) override;
   void visit(Ternary &ternary) override;
   void visit(FieldAccess &acc) override;
   void visit(ArrayAccess &arr) override;
   void visit(Cast &cast) override;
+  void visit(Tuple &tuple) override;
   void visit(ExprStatement &expr) override;
   void visit(AssignMapStatement &assignment) override;
   void visit(AssignVarStatement &assignment) override;
@@ -68,22 +84,35 @@ private:
   bool check_arg(const Call &call, Type type, int arg_num, bool want_literal=false);
   bool check_symbol(const Call &call, int arg_num);
 
-  void check_stack_call(Call &call, Type type);
+  void check_stack_call(Call &call, bool kernel);
   void error(const std::string &msg, const location &loc);
   void warning(const std::string &msg, const location &loc);
 
   void assign_map_type(const Map &map, const SizedType &type);
 
+  void builtin_args_tracepoint(AttachPoint *attach_point, Builtin &builtin);
+  ProbeType single_provider_type(void);
+
+  bool in_loop(void)
+  {
+    return loop_depth_ > 0;
+  };
+  void accept_statements(StatementList *stmts);
+
   Probe *probe_;
+  std::string func_;
   std::map<std::string, SizedType> variable_val_;
   std::map<std::string, SizedType> map_val_;
   std::map<std::string, MapKey> map_key_;
   std::map<std::string, ExpressionList> map_args_;
+  std::map<std::string, SizedType> ap_args_;
   std::unordered_set<StackType> needs_stackid_maps_;
+  uint32_t loop_depth_ = 0;
   bool needs_join_map_ = false;
   bool needs_elapsed_map_ = false;
   bool has_begin_probe_ = false;
   bool has_end_probe_ = false;
+  bool has_child_ = false;
 };
 
 } // namespace ast
