@@ -15,17 +15,20 @@
 
 #include "attached_probe.h"
 #include "bpftrace.h"
-#include "utils.h"
-#include "bcc_syms.h"
-#include "bcc_usdt.h"
-#include "bcc_elf.h"
-#include "libbpf.h"
-#include "utils.h"
-#include "list.h"
 #include "disasm.h"
+#include "list.h"
+#include "usdt.h"
+#include <bcc/bcc_elf.h>
+#include <bcc/bcc_syms.h>
+#include <bcc/bcc_usdt.h>
 #include <linux/perf_event.h>
 #include <linux/version.h>
 
+namespace libbpf {
+#undef __BPF_FUNC_MAPPER
+#include "libbpf/bpf.h"
+} // namespace libbpf
+
 namespace bpftrace {
 
 /*
@@ -68,12 +71,33 @@ bpf_prog_type progtype(ProbeType t)
     case ProbeType::software:   return BPF_PROG_TYPE_PERF_EVENT; break;
     case ProbeType::watchpoint: return BPF_PROG_TYPE_PERF_EVENT; break;
     case ProbeType::hardware:   return BPF_PROG_TYPE_PERF_EVENT; break;
+    case ProbeType::kfunc:
+      return static_cast<enum ::bpf_prog_type>(libbpf::BPF_PROG_TYPE_TRACING);
+      break;
+    case ProbeType::kretfunc:
+      return static_cast<enum ::bpf_prog_type>(libbpf::BPF_PROG_TYPE_TRACING);
+      break;
     default:
       std::cerr << "program type not found" << std::endl;
       abort();
   }
 }
 
+std::string progtypeName(bpf_prog_type t)
+{
+  switch (t)
+  {
+    // clang-format off
+    case BPF_PROG_TYPE_KPROBE:     return "BPF_PROG_TYPE_KPROBE";     break;
+    case BPF_PROG_TYPE_TRACEPOINT: return "BPF_PROG_TYPE_TRACEPOINT"; break;
+    case BPF_PROG_TYPE_PERF_EVENT: return "BPF_PROG_TYPE_PERF_EVENT"; break;
+    // clang-format on
+    default:
+      std::cerr << "invalid program type: " << t << std::endl;
+      abort();
+  }
+}
+
 void check_banned_kretprobes(std::string const& kprobe_name) {
   if (banned_kretprobes.find(kprobe_name) != banned_kretprobes.end()) {
     std::cerr << "error: kretprobe:" << kprobe_name << " can't be used as it might lock up your system." << std::endl;
@@ -81,6 +105,35 @@ void check_banned_kretprobes(std::string const& kprobe_name) {
   }
 }
 
+#ifdef HAVE_BCC_KFUNC
+void AttachedProbe::attach_kfunc(void)
+{
+  tracing_fd_ = bpf_attach_kfunc(progfd_);
+  if (tracing_fd_ < 0)
+    throw std::runtime_error("Error attaching probe: " + probe_.name);
+}
+
+int AttachedProbe::detach_kfunc(void)
+{
+  close(tracing_fd_);
+  return 0;
+}
+#else
+void AttachedProbe::attach_kfunc(void)
+{
+  throw std::runtime_error(
+      "Error attaching probe: " + probe_.name +
+      ", kfunc not available for your linked against bcc version");
+}
+
+int AttachedProbe::detach_kfunc(void)
+{
+  std::cerr << "kfunc not available for linked against bcc version"
+            << std::endl;
+  return -1;
+}
+#endif // HAVE_BCC_KFUNC
+
 AttachedProbe::AttachedProbe(Probe &probe, std::tuple<uint8_t *, uintptr_t> func, bool safe_mode)
   : probe_(probe), func_(func)
 {
@@ -115,6 +168,10 @@ AttachedProbe::AttachedProbe(Probe &probe, std::tuple<uint8_t *, uintptr_t> func
     case ProbeType::hardware:
       attach_hardware();
       break;
+    case ProbeType::kfunc:
+    case ProbeType::kretfunc:
+      attach_kfunc();
+      break;
     default:
       std::cerr << "invalid attached probe type \"" << probetypeName(probe_.type) << "\"" << std::endl;
       abort();
@@ -141,9 +198,6 @@ AttachedProbe::AttachedProbe(Probe &probe, std::tuple<uint8_t *, uintptr_t> func
 
 AttachedProbe::~AttachedProbe()
 {
-  if (progfd_ >= 0)
-    close(progfd_);
-
   int err = 0;
   for (int perf_event_fd : perf_event_fds_)
   {
@@ -159,9 +213,15 @@ AttachedProbe::~AttachedProbe()
     case ProbeType::kretprobe:
       err = bpf_detach_kprobe(eventname().c_str());
       break;
+    case ProbeType::kfunc:
+    case ProbeType::kretfunc:
+      err = detach_kfunc();
+      break;
     case ProbeType::uprobe:
     case ProbeType::uretprobe:
     case ProbeType::usdt:
+      if (usdt_destructor_)
+        usdt_destructor_();
       err = bpf_detach_uprobe(eventname().c_str());
       break;
     case ProbeType::tracepoint:
@@ -179,6 +239,9 @@ AttachedProbe::~AttachedProbe()
   }
   if (err)
     std::cerr << "Error detaching probe: " << probe_.name << std::endl;
+
+  if (progfd_ >= 0)
+    close(progfd_);
 }
 
 std::string AttachedProbe::eventprefix() const
@@ -343,14 +406,29 @@ void AttachedProbe::resolve_offset_uprobe(bool safe_mode)
     sym.address = probe_.address;
     bcc_elf_foreach_sym(probe_.path.c_str(), sym_address_cb, &option, &sym);
 
+    if (!sym.start)
+    {
+      if (safe_mode)
+      {
+        std::stringstream ss;
+        ss << "0x" << std::hex << probe_.address;
+        throw std::runtime_error("Could not resolve address: " + probe_.path +
+                                 ":" + ss.str());
+      }
+      else
+      {
+        std::cerr << "WARNING: could not determine instruction boundary for "
+                  << probe_.name
+                  << " (binary appears stripped). Misaligned probes "
+                     "can lead to tracee crashes!"
+                  << std::endl;
+        offset_ = probe_.address;
+        return;
+      }
+    }
+
     symbol = sym.name;
     func_offset = probe_.address - sym.start;
-
-    if (!sym.start) {
-      std::stringstream ss;
-      ss << "0x" << std::hex << probe_.address;
-      throw std::runtime_error("Could not resolve address: " + probe_.path + ":" + ss.str());
-    }
   }
   else
   {
@@ -455,7 +533,7 @@ void AttachedProbe::resolve_offset_kprobe(bool safe_mode)
 #ifdef HAVE_UNSAFE_PROBE
       buf << " Use --unsafe to skip the userspace check.";
 #else
-      buf << " Compile bpftrace with ALLOW_UNAFE_PROBE option to force skip "
+      buf << " Compile bpftrace with ALLOW_UNSAFE_PROBE option to force skip "
              "the check.";
 #endif
       throw std::runtime_error(buf.str());
@@ -576,9 +654,12 @@ void AttachedProbe::load_prog()
   int prog_len = std::get<1>(func_);
   const char *license = "GPL";
   int log_level = 0;
-  char log_buf[probe_.log_size];
-  char name[STRING_SIZE], *namep;
-  unsigned log_buf_size = sizeof (log_buf);
+
+  uint64_t log_buf_size = probe_.log_size;
+  auto log_buf = std::make_unique<char[]>(log_buf_size);
+  char name[STRING_SIZE];
+  const char *namep;
+  std::string tracing_type, tracing_name;
 
   {
     // Redirect stderr, so we don't get error messages from BCC
@@ -598,6 +679,17 @@ void AttachedProbe::load_prog()
     if (strrchr(name, ':') != NULL)
       namep = strrchr(name, ':') + 1;
 
+    // The bcc_prog_load function now recognizes 'kfunc__/kretfunc__'
+    // prefixes and detects and fills in all the necessary BTF related
+    // attributes for loading the kfunc program.
+
+    tracing_type = probetypeName(probe_.type);
+    if (!tracing_type.empty())
+    {
+      tracing_name = tracing_type + "__" + namep;
+      namep = tracing_name.c_str();
+    }
+
     for (int attempt = 0; attempt < 3; attempt++)
     {
       auto version = kernel_version(attempt);
@@ -622,7 +714,7 @@ void AttachedProbe::load_prog()
                               license,
                               version,
                               log_level,
-                              log_buf,
+                              log_buf.get(),
                               log_buf_size);
       if (progfd_ >= 0)
         break;
@@ -631,7 +723,9 @@ void AttachedProbe::load_prog()
 
   if (progfd_ < 0) {
     if (bt_verbose) {
-      std::cerr << std::endl << "Error log: " << std::endl << log_buf << std::endl;
+      std::cerr << std::endl
+                << "Error log: " << std::endl
+                << log_buf.get() << std::endl;
       if (errno == ENOSPC) {
         std::stringstream errmsg;
         errmsg << "Error: Failed to load program, verification log buffer "
@@ -654,30 +748,29 @@ void AttachedProbe::load_prog()
     if (ret == 0) {
       std::cout << std::endl << "Program ID: " << info.id << std::endl;
     }
-    std::cout << std::endl << "Bytecode: " << std::endl << log_buf << std::endl;
+    std::cout << std::endl
+              << "Bytecode: " << std::endl
+              << log_buf.get() << std::endl;
   }
 }
 
-// XXX(mmarchini): bcc changed the signature of bpf_attach_kprobe, adding a new
-// int parameter at the end. Since there's no reliable way to feature-detect
-// this, we create a function pointer with the long signature and cast
-// bpf_attach_kprobe to this function pointer. If we're on an older bcc
-// version, bpf_attach_kprobe call will be augmented with an extra register
-// being used for the last parameter, even though this register won't be used
-// inside the function. Since the register won't be used this is kinda safe,
-// although not ideal.
-typedef int (*attach_probe_wrapper_signature)(int, enum bpf_probe_attach_type, const char*, const char*, uint64_t, int);
-
 void AttachedProbe::attach_kprobe(bool safe_mode)
 {
   resolve_offset_kprobe(safe_mode);
-  int perf_event_fd = cast_signature<attach_probe_wrapper_signature>(
-      &bpf_attach_kprobe)(progfd_,
-                          attachtype(probe_.type),
-                          eventname().c_str(),
-                          probe_.attach_point.c_str(),
-                          offset_,
-                          0);
+#ifdef LIBBCC_ATTACH_KPROBE_SIX_ARGS_SIGNATURE
+  int perf_event_fd = bpf_attach_kprobe(progfd_,
+                                        attachtype(probe_.type),
+                                        eventname().c_str(),
+                                        probe_.attach_point.c_str(),
+                                        offset_,
+                                        0);
+#else
+  int perf_event_fd = bpf_attach_kprobe(progfd_,
+                                        attachtype(probe_.type),
+                                        eventname().c_str(),
+                                        probe_.attach_point.c_str(),
+                                        offset_);
+#endif
 
   if (perf_event_fd < 0) {
     if (probe_.orig_name != probe_.name) {
@@ -697,12 +790,23 @@ void AttachedProbe::attach_uprobe(bool safe_mode)
 {
   resolve_offset_uprobe(safe_mode);
 
-  int perf_event_fd = bpf_attach_uprobe(progfd_,
-                                        attachtype(probe_.type),
-                                        eventname().c_str(),
-                                        probe_.path.c_str(),
-                                        offset_,
-                                        probe_.pid);
+  int perf_event_fd =
+#ifdef LIBBCC_ATTACH_UPROBE_SEVEN_ARGS_SIGNATURE
+      bpf_attach_uprobe(progfd_,
+                        attachtype(probe_.type),
+                        eventname().c_str(),
+                        probe_.path.c_str(),
+                        offset_,
+                        probe_.pid,
+                        0);
+#else
+      bpf_attach_uprobe(progfd_,
+                        attachtype(probe_.type),
+                        eventname().c_str(),
+                        probe_.path.c_str(),
+                        offset_,
+                        probe_.pid);
+#endif // LIBBCC_ATTACH_UPROBE_SEVEN_ARGS_SIGNATURE
 
   if (perf_event_fd < 0)
     throw std::runtime_error("Error attaching probe: " + probe_.name);
@@ -715,13 +819,44 @@ void AttachedProbe::attach_usdt(int pid)
   struct bcc_usdt_location loc = {};
   int err;
   void *ctx;
+  // TODO: fn_name may need a unique suffix for each attachment on the same
+  // probe:
+  std::string fn_name = "probe_" + probe_.attach_point + "_1";
+
+#ifdef HAVE_BCC_USDT_ADDSEM
+  // NB: we are careful to capture by value here everything that will not
+  // be available in AttachedProbe destructor.
+  auto addsem = [this, fn_name](void *c, int16_t val) -> int {
+    if (this->probe_.ns == "")
+      return bcc_usdt_addsem_probe(
+          c, this->probe_.attach_point.c_str(), fn_name.c_str(), val);
+    else
+      return bcc_usdt_addsem_fully_specified_probe(
+          c,
+          this->probe_.ns.c_str(),
+          this->probe_.attach_point.c_str(),
+          fn_name.c_str(),
+          val);
+  };
+#endif // HAVE_BCC_USDT_ADDSEM
 
   if (pid)
   {
-    //FIXME when iovisor/bcc#2604 is merged, optionally pass probe_.path
+    // FIXME when iovisor/bcc#2064 is merged, optionally pass probe_.path
     ctx = bcc_usdt_new_frompid(pid, nullptr);
     if (!ctx)
       throw std::runtime_error("Error initializing context for probe: " + probe_.name + ", for PID: " + std::to_string(pid));
+
+#ifdef HAVE_BCC_USDT_ADDSEM
+    usdt_destructor_ = [pid, addsem]() {
+      void *c = bcc_usdt_new_frompid(pid, nullptr);
+      if (!c)
+        return;
+
+      addsem(c, -1);
+      bcc_usdt_close(c);
+    };
+#endif // HAVE_BCC_USDT_ADDSEM
   }
   else
   {
@@ -730,10 +865,21 @@ void AttachedProbe::attach_usdt(int pid)
       throw std::runtime_error("Error initializing context for probe: " + probe_.name);
   }
 
-  // TODO: fn_name may need a unique suffix for each attachment on the same probe:
-  std::string fn_name = "probe_" + probe_.attach_point + "_1";
-// see https://github.com/iovisor/bcc/pull/2294 for BCC_USDT_HAS_FULLY_SPECIFIED_PROBE
-#ifdef BCC_USDT_HAS_FULLY_SPECIFIED_PROBE
+#ifndef HAVE_BCC_USDT_ADDSEM
+  // Defer context destruction until probes are detached b/c context
+  // destruction will decrement usdt semaphore count.
+  usdt_destructor_ = [ctx]() { bcc_usdt_close(ctx); };
+#endif // HAVE_BCC_USDT_ADDSEM
+
+#ifdef HAVE_BCC_USDT_ADDSEM
+  // Use semaphore increment API to avoid having to hold onto the usdt context
+  // for the entire tracing session. Reason we do it this way instead of
+  // holding onto usdt context is b/c each usdt context can take lots of memory
+  // (~10MB). This, coupled with --usdt-file-activation and tracees that have a
+  // forking model can cause bpftrace to use huge amounts of memory if we hold
+  // onto the contexts.
+  err = addsem(ctx, +1);
+#elif defined(BCC_USDT_HAS_FULLY_SPECIFIED_PROBE)
   if (probe_.ns == "")
     err = bcc_usdt_enable_probe(ctx, probe_.attach_point.c_str(), fn_name.c_str());
   else
@@ -743,20 +889,49 @@ void AttachedProbe::attach_usdt(int pid)
 #endif
 
   if (err)
-    throw std::runtime_error("Error finding or enabling probe: " + probe_.name);
+  {
+    std::string err;
+    err += "Error finding or enabling probe: " + probe_.name;
+    err += '\n';
+    err += "Try using -p or --usdt-file-activation if there's USDT semaphores";
+    throw std::runtime_error(err);
+  }
 
   auto u = USDTHelper::find(pid, probe_.path, probe_.ns, probe_.attach_point);
-  probe_.path = std::get<USDT_PATH_INDEX>(u);
+  if (!u.has_value())
+    throw std::runtime_error("Failed to find usdt probe: " + eventname());
+  probe_.path = u->path;
 
   err = bcc_usdt_get_location(ctx, probe_.ns.c_str(), probe_.attach_point.c_str(), 0, &loc);
   if (err)
     throw std::runtime_error("Error finding location for probe: " + probe_.name);
   probe_.loc = loc.address;
 
+#ifdef HAVE_BCC_USDT_ADDSEM
+  // If we use the bcc_usdt_addsem*() API, bcc won't decrement semaphore count
+  // in bcc_usdt_close(). So we are free to close context here.
+  bcc_usdt_close(ctx);
+#endif // HAVE_BCC_USDT_ADDSEM
+
   offset_ = resolve_offset(probe_.path, probe_.attach_point, probe_.loc);
 
-  int perf_event_fd = bpf_attach_uprobe(progfd_, attachtype(probe_.type),
-      eventname().c_str(), probe_.path.c_str(), offset_, pid == 0 ? -1 : pid);
+  int perf_event_fd =
+#ifdef LIBBCC_ATTACH_UPROBE_SEVEN_ARGS_SIGNATURE
+      bpf_attach_uprobe(progfd_,
+                        attachtype(probe_.type),
+                        eventname().c_str(),
+                        probe_.path.c_str(),
+                        offset_,
+                        pid == 0 ? -1 : pid,
+                        0);
+#else
+      bpf_attach_uprobe(progfd_,
+                        attachtype(probe_.type),
+                        eventname().c_str(),
+                        probe_.path.c_str(),
+                        offset_,
+                        pid == 0 ? -1 : pid);
+#endif // LIBBCC_ATTACH_UPROBE_SEVEN_ARGS_SIGNATURE
 
   if (perf_event_fd < 0)
   {
@@ -831,7 +1006,7 @@ void AttachedProbe::attach_interval()
   int group_fd = -1;
   int cpu = 0;
 
-  uint64_t period;
+  uint64_t period = 0, freq = 0;
   if (probe_.path == "s")
   {
     period = probe_.freq * 1e9;
@@ -840,14 +1015,28 @@ void AttachedProbe::attach_interval()
   {
     period = probe_.freq * 1e6;
   }
+  else if (probe_.path == "us")
+  {
+    period = probe_.freq * 1e3;
+  }
+  else if (probe_.path == "hz")
+  {
+    freq = probe_.freq;
+  }
   else
   {
     std::cerr << "invalid interval path \"" << probe_.path << "\"" << std::endl;
     abort();
   }
 
-  int perf_event_fd = bpf_attach_perf_event(progfd_, PERF_TYPE_SOFTWARE,
-      PERF_COUNT_SW_CPU_CLOCK, period, 0, pid, cpu, group_fd);
+  int perf_event_fd = bpf_attach_perf_event(progfd_,
+                                            PERF_TYPE_SOFTWARE,
+                                            PERF_COUNT_SW_CPU_CLOCK,
+                                            period,
+                                            freq,
+                                            pid,
+                                            cpu,
+                                            group_fd);
 
   if (perf_event_fd < 0)
     throw std::runtime_error("Error attaching probe: " + probe_.name);
@@ -946,7 +1135,7 @@ void AttachedProbe::attach_watchpoint(int pid, const std::string& mode)
       attr.bp_type |= HW_BREAKPOINT_X;
   }
 
-  attr.bp_addr = probe_.addr;
+  attr.bp_addr = probe_.address;
   attr.bp_len = probe_.len;
   // Generate a notification every 1 event; we care about every event
   attr.sample_period = 1;
