@@ -1,17 +1,19 @@
 #pragma once
 
+#include <functional>
 #include <string>
 #include <tuple>
 #include <vector>
 
 #include "types.h"
 
-#include "libbpf.h"
+#include <bcc/libbpf.h>
 
 namespace bpftrace {
 
 bpf_probe_attach_type attachtype(ProbeType t);
 bpf_prog_type progtype(ProbeType t);
+std::string progtypeName(bpf_prog_type t);
 
 class AttachedProbe
 {
@@ -40,12 +42,18 @@ private:
   void attach_software();
   void attach_hardware();
   void attach_watchpoint(int pid, const std::string &mode);
+  void attach_kfunc(void);
+  int detach_kfunc(void);
 
   Probe &probe_;
   std::tuple<uint8_t *, uintptr_t> func_;
   std::vector<int> perf_event_fds_;
   int progfd_ = -1;
   uint64_t offset_ = 0;
+#ifdef HAVE_BCC_KFUNC
+  int tracing_fd_ = -1;
+#endif
+  std::function<void()> usdt_destructor_;
 };
 
 } // namespace bpftrace
