@@ -25,7 +25,7 @@ public:
   void error(std::ostream &, const location &, const std::string &);
   void error(const location &l, const std::string &m);
   void error(const std::string &m);
-  ast::Program *root_;
+  ast::Program *root_{ nullptr };
 
   BPFtrace &bpftrace_;
 
