@@ -11,13 +11,16 @@
 #include "bpffeature.h"
 #include "bpforc.h"
 #include "bpftrace.h"
+#include "child.h"
 #include "clang_parser.h"
 #include "codegen_llvm.h"
 #include "driver.h"
 #include "field_analyser.h"
 #include "list.h"
+#include "lockdown.h"
 #include "output.h"
 #include "printer.h"
+#include "procmon.h"
 #include "semantic_analyser.h"
 #include "tracepoint_format_parser.h"
 
@@ -37,7 +40,9 @@ void usage()
   // clang-format off
   std::cerr << "USAGE:" << std::endl;
   std::cerr << "    bpftrace [options] filename" << std::endl;
-  std::cerr << "    bpftrace [options] -e 'program'" << std::endl << std::endl;
+  std::cerr << "    bpftrace [options] - <stdin input>" << std::endl;
+  std::cerr << "    bpftrace [options] -e 'program'" << std::endl;
+  std::cerr << std::endl;
   std::cerr << "OPTIONS:" << std::endl;
   std::cerr << "    -B MODE        output buffering mode ('full', 'none')" << std::endl;
   std::cerr << "    -f FORMAT      output format ('text', 'json')" << std::endl;
@@ -52,20 +57,26 @@ void usage()
   std::cerr << "    -l [search]    list probes" << std::endl;
   std::cerr << "    -p PID         enable USDT probes on PID" << std::endl;
   std::cerr << "    -c 'CMD'       run CMD and enable USDT probes on resulting process" << std::endl;
+  std::cerr << "    --usdt-file-activation" << std::endl;
+  std::cerr << "                   activate usdt semaphores based on file path" << std::endl;
   std::cerr << "    --unsafe       allow unsafe builtin functions" << std::endl;
   std::cerr << "    -v             verbose messages" << std::endl;
   std::cerr << "    --info         Print information about kernel BPF support" << std::endl;
+  std::cerr << "    -k             emit a warning when a bpf helper returns an error (except read functions)" << std::endl;
+  std::cerr << "    -kk            check all bpf helper functions" << std::endl;
   std::cerr << "    -V, --version  bpftrace version" << std::endl << std::endl;
   std::cerr << "ENVIRONMENT:" << std::endl;
-  std::cerr << "    BPFTRACE_STRLEN           [default: 64] bytes on BPF stack per str()" << std::endl;
-  std::cerr << "    BPFTRACE_NO_CPP_DEMANGLE  [default: 0] disable C++ symbol demangling" << std::endl;
-  std::cerr << "    BPFTRACE_MAP_KEYS_MAX     [default: 4096] max keys in a map" << std::endl;
-  std::cerr << "    BPFTRACE_CAT_BYTES_MAX    [default: 10k] maximum bytes read by cat builtin" << std::endl;
-  std::cerr << "    BPFTRACE_MAX_PROBES       [default: 512] max number of probes" << std::endl;
-  std::cerr << "    BPFTRACE_LOG_SIZE         [default: 409600] log size in bytes" << std::endl;
-  std::cerr << "    BPFTRACE_NO_USER_SYMBOLS  [default: 0] disable user symbol resolution" << std::endl;
-  std::cerr << "    BPFTRACE_VMLINUX          [default: None] vmlinux path used for kernel symbol resolution" << std::endl;
-  std::cerr << "    BPFTRACE_BTF              [default: None] BTF file" << std::endl;
+  std::cerr << "    BPFTRACE_STRLEN             [default: 64] bytes on BPF stack per str()" << std::endl;
+  std::cerr << "    BPFTRACE_NO_CPP_DEMANGLE    [default: 0] disable C++ symbol demangling" << std::endl;
+  std::cerr << "    BPFTRACE_MAP_KEYS_MAX       [default: 4096] max keys in a map" << std::endl;
+  std::cerr << "    BPFTRACE_CAT_BYTES_MAX      [default: 10k] maximum bytes read by cat builtin" << std::endl;
+  std::cerr << "    BPFTRACE_MAX_PROBES         [default: 512] max number of probes" << std::endl;
+  std::cerr << "    BPFTRACE_LOG_SIZE           [default: 1000000] log size in bytes" << std::endl;
+  std::cerr << "    BPFTRACE_PERF_RB_PAGES      [default: 64] pages per CPU to allocate for ring buffer" << std::endl;
+  std::cerr << "    BPFTRACE_NO_USER_SYMBOLS    [default: 0] disable user symbol resolution" << std::endl;
+  std::cerr << "    BPFTRACE_CACHE_USER_SYMBOLS [default: auto] enable user symbol cache" << std::endl;
+  std::cerr << "    BPFTRACE_VMLINUX            [default: none] vmlinux path used for kernel symbol resolution" << std::endl;
+  std::cerr << "    BPFTRACE_BTF                [default: none] BTF file" << std::endl;
   std::cerr << std::endl;
   std::cerr << "EXAMPLES:" << std::endl;
   std::cerr << "bpftrace -l '*sleep*'" << std::endl;
@@ -90,30 +101,6 @@ static void enforce_infinite_rlimit() {
         "\"ulimit -l 8192\" to fix the problem" << std::endl;
 }
 
-#ifdef BUILD_ASAN
-static void cap_memory_limits()
-{
-}
-#else
-static void cap_memory_limits() {
-  struct rlimit rl = {};
-  int err;
-  uint64_t memory_limit_bytes = 1 * 1024 * 1024 * 1024;
-
-  // this is a safety measure for issue #528 "LLVM ERROR: out of memory",
-  // and caps bpftrace memory to 1 Gbyte. This may be removed once the LLVM
-  // issue has been fixed, and this is no longer deemed necessary.
-  rl.rlim_max = memory_limit_bytes;
-  rl.rlim_cur = rl.rlim_max;
-  err = setrlimit(RLIMIT_AS, &rl);
-  err += setrlimit(RLIMIT_RSS, &rl);
-  if (err)
-    std::cerr << std::strerror(err)<<": couldn't set RLIMIT_AS and " <<
-        "RLIMIT_RSS for bpftrace (these are a temporary precaution to stop " <<
-        "accidental large program loads, and are not required" << std::endl;
-}
-#endif // BUILD_ASAN
-
 bool is_root()
 {
   if (geteuid() != 0)
@@ -125,22 +112,101 @@ bool is_root()
     return true;
 }
 
+static int info()
+{
+  struct utsname utsname;
+  uname(&utsname);
+
+  std::cerr << "System" << std::endl
+            << "  OS: " << utsname.sysname << " " << utsname.release << " "
+            << utsname.version << std::endl
+            << "  Arch: " << utsname.machine << std::endl;
+
+  std::cerr << std::endl
+            << "Build" << std::endl
+            << "  version: " << BPFTRACE_VERSION << std::endl
+            << "  LLVM: " << LLVM_VERSION_MAJOR << std::endl
+            << "  foreach_sym: "
+#ifdef HAVE_BCC_ELF_FOREACH_SYM
+            << "yes" << std::endl
+#else
+            << "no" << std::endl
+#endif
+            << "  unsafe uprobe: "
+#ifdef HAVE_UNSAFE_UPROBE
+            << "yes" << std::endl;
+#else
+            << "no" << std::endl;
+#endif
+  std::cerr << "  bfd: "
+#ifdef HAVE_BFD_DISASM
+            << "yes" << std::endl;
+#else
+            << "no" << std::endl;
+#endif
+  std::cerr << "  bpf_attach_kfunc: "
+#ifdef HAVE_BCC_KFUNC
+            << "yes" << std::endl;
+#else
+            << "no" << std::endl;
+#endif
+  std::cerr << "  bcc_usdt_addsem: "
+#ifdef HAVE_BCC_USDT_ADDSEM
+            << "yes" << std::endl;
+#else
+            << "no" << std::endl;
+#endif
+  std::cerr << "  bcc bpf_attach_uprobe refcount: "
+#ifdef LIBBCC_ATTACH_UPROBE_SEVEN_ARGS_SIGNATURE
+            << "yes" << std::endl;
+#else
+            << "no" << std::endl;
+#endif
+  std::cerr << "  libbpf: "
+#ifdef HAVE_LIBBPF
+            << "yes" << std::endl;
+#else
+            << "no" << std::endl;
+#endif
+  std::cerr << "  libbpf btf dump: "
+#ifdef HAVE_LIBBPF_BTF_DUMP
+            << "yes" << std::endl;
+#else
+            << "no" << std::endl;
+#endif
+  std::cerr << "  libbpf btf dump type decl: "
+#ifdef HAVE_LIBBPF_BTF_DUMP_EMIT_TYPE_DECL
+            << "yes" << std::endl;
+#else
+            << "no" << std::endl;
+#endif
+
+  std::cerr << std::endl;
+
+  std::cerr << BPFfeature().report();
+
+  return 0;
+}
+
 int main(int argc, char *argv[])
 {
   int err;
-  char *pid_str = nullptr;
-  char *cmd_str = nullptr;
+  std::string pid_str;
+  std::string cmd_str;
   bool listing = false;
   bool safe_mode = true;
   bool force_btf = false;
+  bool usdt_file_activation = false;
+  int helper_check_level = 0;
   std::string script, search, file_name, output_file, output_format;
   OutputBufferConfig obc = OutputBufferConfig::UNSET;
   int c;
 
-  const char* const short_options = "dbB:f:e:hlp:vc:Vo:I:";
+  const char* const short_options = "dbB:f:e:hlp:vc:Vo:I:k";
   option long_options[] = {
     option{ "help", no_argument, nullptr, 'h' },
     option{ "version", no_argument, nullptr, 'V' },
+    option{ "usdt-file-activation", no_argument, nullptr, '$' },
     option{ "unsafe", no_argument, nullptr, 'u' },
     option{ "btf", no_argument, nullptr, 'b' },
     option{ "include", required_argument, nullptr, '#' },
@@ -154,13 +220,11 @@ int main(int argc, char *argv[])
   {
     switch (c)
     {
-      case 2000:
+      case 2000: // --info
         if (is_root())
-        {
-          std::cerr << BPFfeature().report();
-          return 0;
-        }
+          return info();
         return 1;
+        break;
       case 'o':
         output_file = optarg;
         break;
@@ -207,6 +271,9 @@ int main(int argc, char *argv[])
       case 'c':
         cmd_str = optarg;
         break;
+      case '$':
+        usdt_file_activation = true;
+        break;
       case 'u':
         safe_mode = false;
         break;
@@ -219,6 +286,14 @@ int main(int argc, char *argv[])
       case 'V':
         std::cout << "bpftrace " << BPFTRACE_VERSION << std::endl;
         return 0;
+      case 'k':
+        helper_check_level++;
+        if (helper_check_level >= 3)
+        {
+          usage();
+          return 1;
+        }
+        break;
       default:
         usage();
         return 1;
@@ -237,7 +312,7 @@ int main(int argc, char *argv[])
     return 1;
   }
 
-  if (cmd_str && pid_str)
+  if (!cmd_str.empty() && !pid_str.empty())
   {
     std::cerr << "USAGE: Cannot use both -c and -p." << std::endl;
     usage();
@@ -288,21 +363,22 @@ int main(int argc, char *argv[])
   BPFtrace bpftrace(std::move(output));
   Driver driver(bpftrace);
 
+  bpftrace.usdt_file_activation_ = usdt_file_activation;
   bpftrace.safe_mode_ = safe_mode;
   bpftrace.force_btf_ = force_btf;
+  bpftrace.helper_check_level_ = helper_check_level;
 
-  // PID is currently only used for USDT probes that need enabling. Future work:
-  // - make PID a filter for all probe types: pass to perf_event_open(), etc.
-  // - provide PID in USDT probe specification as a way to override -p.
-  bpftrace.pid_ = 0;
-  if (pid_str)
+  if (!pid_str.empty())
   {
-    if (!is_numeric(pid_str))
+    try
     {
-      std::cerr << "ERROR: pid '" << pid_str << "' is not a valid number." << std::endl;
+      bpftrace.procmon_ = std::make_unique<ProcMon>(pid_str);
+    }
+    catch (const std::exception& e)
+    {
+      std::cerr << "ERROR: " << e.what() << std::endl;
       return 1;
     }
-    bpftrace.pid_ = strtol(pid_str, NULL, 10);
   }
 
   // Listing probes
@@ -330,20 +406,34 @@ int main(int argc, char *argv[])
       return 1;
     }
     std::string filename(argv[optind]);
-    std::ifstream file(filename);
-    if (file.fail())
-    {
-      std::cerr << "Error opening file '" << filename << "': ";
-      std::cerr << std::strerror(errno) << std::endl;
-      return -1;
-    }
-
     std::stringstream buf;
-    buf << file.rdbuf();
-    driver.source(filename, buf.str());
-    err = driver.parse();
-    if (err)
-      return err;
+
+    if (filename == "-")
+    {
+      std::string line;
+      while (std::getline(std::cin, line))
+      {
+        // Note we may add an extra newline if the input doesn't end in a new
+        // line. This should not matter because bpftrace (the language) is not
+        // whitespace sensitive.
+        buf << line << std::endl;
+      }
+
+      driver.source("stdin", buf.str());
+    }
+    else
+    {
+      std::ifstream file(filename);
+      if (file.fail())
+      {
+        std::cerr << "Error opening file '" << filename << "': ";
+        std::cerr << std::strerror(errno) << std::endl;
+        return -1;
+      }
+
+      buf << file.rdbuf();
+      driver.source(filename, buf.str());
+    }
 
     optind++;
   }
@@ -351,14 +441,30 @@ int main(int argc, char *argv[])
   {
     // Script is provided as a command line argument
     driver.source("stdin", script);
-    err = driver.parse();
-    if (err)
-      return err;
   }
 
+  // Load positional parameters before driver runs so positional
+  // parameters used inside attach point definitions can be resolved.
+  while (optind < argc)
+  {
+    bpftrace.add_param(argv[optind]);
+    optind++;
+  }
+
+  err = driver.parse();
+  if (err)
+    return err;
+
   if (!is_root())
     return 1;
 
+  auto lockdown_state = lockdown::detect(bpftrace.feature_);
+  if (lockdown_state == lockdown::LockdownState::Confidentiality)
+  {
+    lockdown::emit_warning(std::cerr);
+    return 1;
+  }
+
   ast::FieldAnalyser fields(driver.root_, bpftrace);
   err = fields.analyse();
   if (err)
@@ -368,14 +474,6 @@ int main(int argc, char *argv[])
   // rlimit?
   enforce_infinite_rlimit();
 
-  cap_memory_limits();
-
-  // positional parameters
-  while (optind < argc) {
-    bpftrace.add_param(argv[optind]);
-    optind++;
-  }
-
   // defaults
   bpftrace.join_argnum_ = 16;
   bpftrace.join_argsize_ = 1024;
@@ -418,6 +516,9 @@ int main(int argc, char *argv[])
   if (!get_uint64_env_var("BPFTRACE_LOG_SIZE", bpftrace.log_size_))
     return 1;
 
+  if (!get_uint64_env_var("BPFTRACE_PERF_RB_PAGES", bpftrace.perf_rb_pages_))
+    return 1;
+
   if (const char* env_p = std::getenv("BPFTRACE_CAT_BYTES_MAX"))
   {
     uint64_t proposed;
@@ -443,14 +544,39 @@ int main(int argc, char *argv[])
     }
   }
 
-  if (cmd_str)
+  if (const char* env_p = std::getenv("BPFTRACE_CACHE_USER_SYMBOLS"))
+  {
+    std::string s(env_p);
+    if (s == "1")
+      bpftrace.cache_user_symbols_ = true;
+    else if (s == "0")
+      bpftrace.cache_user_symbols_ = false;
+    else
+    {
+      std::cerr << "Env var 'BPFTRACE_CACHE_USER_SYMBOLS' did not contain a "
+                   "valid value (0 or 1)."
+                << std::endl;
+      return 1;
+    }
+  }
+  else
+  {
+    // enable user symbol cache if ASLR is disabled on system or `-c` option is
+    // given
+    bpftrace.cache_user_symbols_ = !cmd_str.empty() ||
+                                   !bpftrace.is_aslr_enabled(-1);
+  }
+
+  if (!cmd_str.empty())
     bpftrace.cmd_ = cmd_str;
 
-  if (TracepointFormatParser::parse(driver.root_) == false)
+  if (TracepointFormatParser::parse(driver.root_, bpftrace) == false)
     return 1;
 
   if (bt_debug != DebugLevel::kNone)
   {
+    std::cout << "\nAST\n";
+    std::cout << "-------------------\n";
     ast::Printer p(std::cout);
     driver.root_->accept(p);
     std::cout << std::endl;
@@ -497,22 +623,49 @@ int main(int argc, char *argv[])
   if (err)
     return err;
 
-  BPFfeature features;
-
-  ast::SemanticAnalyser semantics(driver.root_, bpftrace, features);
+  ast::SemanticAnalyser semantics(
+      driver.root_, bpftrace, bpftrace.feature_, !cmd_str.empty());
   err = semantics.analyse();
   if (err)
     return err;
 
-  if (bpftrace.has_child_cmd() && (bpftrace.spawn_child() < 0))
-    return 1;
+  if (bt_debug != DebugLevel::kNone)
+  {
+    std::cout << "\nAST after semantic analysis\n";
+    std::cout << "-------------------\n";
+    ast::Printer p(std::cout, true);
+    driver.root_->accept(p);
+    std::cout << std::endl;
+  }
 
   err = semantics.create_maps(bt_debug != DebugLevel::kNone);
   if (err)
     return err;
 
+  if (!cmd_str.empty())
+  {
+    try
+    {
+      bpftrace.child_ = std::make_unique<ChildProc>(cmd_str);
+    }
+    catch (const std::runtime_error& e)
+    {
+      std::cerr << "Failed to fork child: " << e.what() << std::endl;
+      return -1;
+    }
+  }
+
   ast::CodegenLLVM llvm(driver.root_, bpftrace);
-  auto bpforc = llvm.compile(bt_debug);
+  std::unique_ptr<BpfOrc> bpforc;
+  try
+  {
+    bpforc = llvm.compile(bt_debug);
+  }
+  catch (const std::exception& ex)
+  {
+    std::cerr << "Failed to compile: " << ex.what() << std::endl;
+    return 1;
+  }
 
   if (bt_debug != DebugLevel::kNone)
     return 0;
@@ -553,6 +706,16 @@ int main(int argc, char *argv[])
   std::cout << "\n\n";
 
   err = bpftrace.print_maps();
+
+  if (bt_verbose && bpftrace.child_)
+  {
+    auto val = 0;
+    if ((val = bpftrace.child_->term_signal()) > -1)
+      std::cout << "Child terminated by signal: " << val << std::endl;
+    if ((val = bpftrace.child_->exit_code()) > -1)
+      std::cout << "Child exited with code: " << val << std::endl;
+  }
+
   if (err)
     return err;
 
