@@ -3,9 +3,9 @@
 #include <unistd.h>
 #include <linux/version.h>
 
-#include "libbpf.h"
-#include "utils.h"
 #include "bpftrace.h"
+#include "utils.h"
+#include <bcc/libbpf.h>
 
 #include "map.h"
 
@@ -30,20 +30,25 @@ Map::Map(const std::string &name, const SizedType &type, const MapKey &key, int
   lqstep = step;
 
   int key_size = key.size();
-  if (type.type == Type::hist || type.type == Type::lhist ||
-      type.type == Type::avg || type.type == Type::stats)
+  if (type.IsHistTy() || type.IsLhistTy() || type.IsAvgTy() || type.IsStatsTy())
     key_size += 8;
   if (key_size == 0)
     key_size = 8;
 
-  if ((type.type == Type::hist || type.type == Type::lhist || type.type == Type::count ||
-      type.type == Type::sum || type.type == Type::min || type.type == Type::max ||
-      type.type == Type::avg || type.type == Type::stats) &&
-      (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 6, 0)))
+  if (type.IsCountTy() && !key.args_.size())
+  {
+    map_type_ = BPF_MAP_TYPE_PERCPU_ARRAY;
+    max_entries = 1;
+    key_size = 4;
+  }
+  else if ((type.IsHistTy() || type.IsLhistTy() || type.IsCountTy() ||
+            type.IsSumTy() || type.IsMinTy() || type.IsMaxTy() ||
+            type.IsAvgTy() || type.IsStatsTy()) &&
+           (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 6, 0)))
   {
       map_type_ = BPF_MAP_TYPE_PERCPU_HASH;
   }
-  else if (type.type == Type::join)
+  else if (type.IsJoinTy())
   {
     map_type_ = BPF_MAP_TYPE_PERCPU_ARRAY;
     max_entries = 1;
