@@ -13,9 +13,8 @@
 #include <tuple>
 #include <unistd.h>
 
-#include "bcc_elf.h"
-#include "bcc_usdt.h"
 #include "utils.h"
+#include <bcc/bcc_elf.h>
 
 namespace {
 
@@ -88,138 +87,32 @@ const struct vmlinux_location vmlinux_locs[] = {
   { nullptr, false },
 };
 
-static bool provider_cache_loaded = false;
-
-// Maps all providers of pid to vector of tracepoints on that provider
-static std::map<std::string, usdt_probe_list> usdt_provider_cache;
-
 static bool pid_in_different_mountns(int pid);
 static std::vector<std::string>
 resolve_binary_path(const std::string &cmd, const char *env_paths, int pid);
 
-static void usdt_probe_each(struct bcc_usdt *usdt_probe)
+void StdioSilencer::silence()
 {
-  usdt_provider_cache[usdt_probe->provider].push_back(std::make_tuple(usdt_probe->bin_path, usdt_probe->provider, usdt_probe->name));
+  fflush(ofile);
+  int fd = fileno(ofile);
+  old_stdio_ = dup(fd);
+  int new_stdio_ = open("/dev/null", O_WRONLY);
+  dup2(new_stdio_, fd);
+  close(new_stdio_);
 }
 
-void StderrSilencer::silence()
+StdioSilencer::~StdioSilencer()
 {
-  fflush(stderr);
-  old_stderr_ = dup(STDERR_FILENO);
-  int new_stderr_ = open("/dev/null", O_WRONLY);
-  dup2(new_stderr_, STDERR_FILENO);
-  close(new_stderr_);
-}
-
-StderrSilencer::~StderrSilencer()
-{
-  if (old_stderr_ != -1)
+  if (old_stdio_ != -1)
   {
-    fflush(stderr);
-    dup2(old_stderr_, STDERR_FILENO);
-    close(old_stderr_);
-    old_stderr_ = -1;
+    fflush(ofile);
+    int fd = fileno(ofile);
+    dup2(old_stdio_, fd);
+    close(old_stdio_);
+    old_stdio_ = -1;
   }
 }
 
-usdt_probe_entry USDTHelper::find(
-    int pid,
-    const std::string &target,
-    const std::string &provider,
-    const std::string &name)
-{
-
-  if (pid > 0)
-    read_probes_for_pid(pid);
-  else
-    read_probes_for_path(target);
-
-  usdt_probe_list probes = usdt_provider_cache[provider];
-
-  auto it = std::find_if(probes.begin(), probes.end(), [&name](const usdt_probe_entry& e) {return std::get<USDT_FNAME_INDEX>(e) == name;});
-  if (it != probes.end()) {
-    return *it;
-  } else {
-    return std::make_tuple("", "", "");
-  }
-}
-
-usdt_probe_list USDTHelper::probes_for_provider(const std::string &provider)
-{
-  usdt_probe_list probes;
-
-  if(!provider_cache_loaded) {
-    std::cerr << "cannot read probes by provider before providers have been loaded by pid or path." << std::endl;
-    return probes;
-  }
-
-  read_probes_for_pid(0);
-  return usdt_provider_cache[provider];
-}
-
-usdt_probe_list USDTHelper::probes_for_pid(int pid)
-{
-  read_probes_for_pid(pid);
-
-  usdt_probe_list probes;
-  for (auto const& usdt_probes : usdt_provider_cache)
-  {
-    probes.insert( probes.end(), usdt_probes.second.begin(), usdt_probes.second.end() );
-  }
-  return probes;
-}
-
-usdt_probe_list USDTHelper::probes_for_path(const std::string &path)
-{
-  read_probes_for_path(path);
-
-  usdt_probe_list probes;
-  for (auto const& usdt_probes : usdt_provider_cache)
-  {
-    probes.insert( probes.end(), usdt_probes.second.begin(), usdt_probes.second.end() );
-  }
-  return probes;
-}
-
-void USDTHelper::read_probes_for_pid(int pid)
-{
-  if(provider_cache_loaded)
-    return;
-
-  if (pid > 0) {
-    void *ctx = bcc_usdt_new_frompid(pid, nullptr);
-    if (ctx == nullptr) {
-      std::cerr << "failed to initialize usdt context for pid: " << pid << std::endl;
-      if (kill(pid, 0) == -1 && errno == ESRCH) {
-        std::cerr << "hint: process not running" << std::endl;
-      }
-      return;
-    }
-    bcc_usdt_foreach(ctx, usdt_probe_each);
-    bcc_usdt_close(ctx);
-
-    provider_cache_loaded = true;
-  } else {
-    std::cerr << "a pid must be specified to list USDT probes by PID" << std::endl;
-  }
-}
-
-void USDTHelper::read_probes_for_path(const std::string &path)
-{
-  if(provider_cache_loaded)
-    return;
-
-  void *ctx = bcc_usdt_new_frompath(path.c_str());
-  if (ctx == nullptr) {
-    std::cerr << "failed to initialize usdt context for path " << path << std::endl;
-    return;
-  }
-  bcc_usdt_foreach(ctx, usdt_probe_each);
-  bcc_usdt_close(ctx);
-
-  provider_cache_loaded = true;
-}
-
 bool get_uint64_env_var(const std::string &str, uint64_t &dest)
 {
   if (const char* env_p = std::getenv(str.c_str()))
@@ -258,11 +151,17 @@ bool has_wildcard(const std::string &str)
           str.find("]") != std::string::npos);
 }
 
-std::vector<std::string> split_string(const std::string &str, char delimiter) {
+std::vector<std::string> split_string(const std::string &str,
+                                      char delimiter,
+                                      bool remove_empty)
+{
   std::vector<std::string> elems;
   std::stringstream ss(str);
   std::string value;
   while(std::getline(ss, value, delimiter)) {
+    if (remove_empty && value.empty())
+      continue;
+
     elems.push_back(value);
   }
   return elems;
@@ -357,7 +256,7 @@ std::vector<std::string> get_kernel_cflags(
   cflags.push_back("-D__HAVE_BUILTIN_BSWAP16__");
   cflags.push_back("-D__HAVE_BUILTIN_BSWAP32__");
   cflags.push_back("-D__HAVE_BUILTIN_BSWAP64__");
-  cflags.push_back("-DKBUILD_MODNAME='\"bpftrace\"'");
+  cflags.push_back("-DKBUILD_MODNAME=\"bpftrace\"");
 
   // If ARCH env variable is set, pass this along.
   if (archenv)
@@ -421,6 +320,9 @@ namespace {
     }
 
     if (::stat("/sys/kernel/kheaders.tar.xz", &stat_buf) != 0) {
+      StderrSilencer silencer;
+      silencer.silence();
+
       FILE* modprobe = ::popen("modprobe kheaders", "w");
       if (modprobe == nullptr || pclose(modprobe) != 0) {
         return "";
@@ -535,6 +437,13 @@ bool is_unsafe_func(const std::string &func_name)
       });
 }
 
+bool is_compile_time_func(const std::string &func_name)
+{
+  return std::any_of(COMPILE_TIME_FUNCS.begin(),
+                     COMPILE_TIME_FUNCS.end(),
+                     [&](const auto &cand) { return func_name == cand; });
+}
+
 std::string exec_system(const char* cmd)
 {
   std::array<char, 128> buffer;
@@ -628,9 +537,20 @@ resolve_binary_path(const std::string &cmd, const char *env_paths, int pid)
 std::string path_for_pid_mountns(int pid, const std::string &path)
 {
   std::ostringstream pid_relative_path;
-  std::string root =
-      (path.length() >= 1 && path.at(0) == '/') ? "/root" : "/root/";
-  pid_relative_path << "/proc/" << pid << root << path;
+  char pid_root[64];
+
+  snprintf(pid_root, sizeof(pid_root), "/proc/%d/root", pid);
+
+  if (path.find(pid_root) != 0)
+  {
+    std::string sep = (path.length() >= 1 && path.at(0) == '/') ? "" : "/";
+    pid_relative_path << pid_root << sep << path;
+  }
+  else
+  {
+    // The path is already relative to the pid's root
+    pid_relative_path << path;
+  }
   return pid_relative_path.str();
 }
 
@@ -756,7 +676,67 @@ std::string str_join(const std::vector<std::string> &list, const std::string &de
 
 bool is_numeric(const std::string &s)
 {
-  return !s.empty() && std::all_of(s.begin(), s.end(), ::isdigit);
+  std::size_t idx;
+  try
+  {
+    std::stoll(s, &idx, 0);
+  }
+  catch (...)
+  {
+    return false;
+  }
+  return idx == s.size();
+}
+
+bool symbol_has_cpp_mangled_signature(const std::string &sym_name)
+{
+  if (!sym_name.rfind("_Z", 0) || !sym_name.rfind("____Z", 0))
+    return true;
+  else
+    return false;
+}
+
+pid_t parse_pid(const std::string &str)
+{
+  try
+  {
+    constexpr ssize_t pid_max = 4 * 1024 * 1024;
+    std::size_t idx = 0;
+    auto pid = std::stol(str, &idx, 10);
+    // Detect cases like `13ABC`
+    if (idx < str.size())
+      throw InvalidPIDException(str, "is not a valid decimal number");
+    if (pid < 1 || pid > pid_max)
+      throw InvalidPIDException(str,
+                                "out of valid pid range [1," +
+                                    std::to_string(pid_max) + "]");
+    return pid;
+  }
+  catch (const std::out_of_range &e)
+  {
+    throw InvalidPIDException(str, "outside of integer range");
+  }
+  catch (const std::invalid_argument &e)
+  {
+    throw InvalidPIDException(str, "is not a valid decimal number");
+  }
+}
+
+std::string hex_format_buffer(const char *buf, size_t size)
+{
+  // Allow enough space for every byte to be sanitized in the form "\x00"
+  char s[size * 4 + 1];
+
+  size_t offset = 0;
+  for (size_t i = 0; i < size; i++)
+    if (buf[i] >= 32 && buf[i] <= 126)
+      offset += sprintf(s + offset, "%c", ((const uint8_t *)buf)[i]);
+    else
+      offset += sprintf(s + offset, "\\x%02x", ((const uint8_t *)buf)[i]);
+
+  s[offset] = '\0';
+
+  return std::string(s);
 }
 
 } // namespace bpftrace
