@@ -18,16 +18,6 @@ struct vmlinux_location
   bool raw;         // file is either as ELF (false) or raw BTF data (true)
 };
 extern const struct vmlinux_location vmlinux_locs[];
-
-typedef enum _USDT_TUPLE_ORDER_
-{
-  USDT_PATH_INDEX,
-  USDT_PROVIDER_INDEX,
-  USDT_FNAME_INDEX
-} usdt_probe_entry_enum;
-typedef std::tuple<std::string, std::string, std::string> usdt_probe_entry;
-typedef std::vector<usdt_probe_entry> usdt_probe_list;
-
 class MountNSException : public std::exception
 {
 public:
@@ -44,29 +34,53 @@ private:
   std::string msg_;
 };
 
-class StderrSilencer
+class InvalidPIDException : public std::exception
 {
 public:
-  StderrSilencer() = default;
-  ~StderrSilencer();
-  void silence();
+  InvalidPIDException(const std::string &pid, const std::string &msg)
+  {
+    msg_ = "pid '" + pid + "' " + msg;
+  }
+
+  const char *what() const noexcept override
+  {
+    return msg_.c_str();
+  }
 
 private:
-  int old_stderr_ = -1;
+  std::string msg_;
 };
 
-class USDTHelper
+class StdioSilencer
 {
 public:
-  static usdt_probe_entry find(int pid,
-                               const std::string &target,
-                               const std::string &provider,
-                               const std::string &name);
-  static usdt_probe_list probes_for_provider(const std::string &provider);
-  static usdt_probe_list probes_for_pid(int pid);
-  static usdt_probe_list probes_for_path(const std::string &path);
-  static void read_probes_for_pid(int pid);
-  static void read_probes_for_path(const std::string &path);
+  StdioSilencer() = default;
+  ~StdioSilencer();
+  void silence();
+
+protected:
+  FILE *ofile;
+
+private:
+  int old_stdio_ = -1;
+};
+
+class StderrSilencer : public StdioSilencer
+{
+public:
+  StderrSilencer()
+  {
+    ofile = stderr;
+  }
+};
+
+class StdoutSilencer : public StdioSilencer
+{
+public:
+  StdoutSilencer()
+  {
+    ofile = stdout;
+  }
 };
 
 // Hack used to suppress build warning related to #474
@@ -91,8 +105,6 @@ struct DeprecatedName
 
 static std::vector<DeprecatedName> DEPRECATED_LIST =
 {
-  { "stack", "kstack"},
-  { "sym", "ksym"},
 };
 
 static std::vector<std::string> UNSAFE_BUILTIN_FUNCS = {
@@ -101,10 +113,14 @@ static std::vector<std::string> UNSAFE_BUILTIN_FUNCS = {
   "override",
 };
 
+static std::vector<std::string> COMPILE_TIME_FUNCS = { "cgroupid" };
+
 bool get_uint64_env_var(const ::std::string &str, uint64_t &dest);
 std::string get_pid_exe(pid_t pid);
 bool has_wildcard(const std::string &str);
-std::vector<std::string> split_string(const std::string &str, char delimiter);
+std::vector<std::string> split_string(const std::string &str,
+                                      char delimiter,
+                                      bool remove_empty = false);
 bool wildcard_match(const std::string &str,
                     std::vector<std::string> &tokens,
                     bool start_wildcard,
@@ -119,6 +135,7 @@ std::vector<std::string> get_kernel_cflags(const char *uname_machine,
                                            const std::string &kobj);
 const std::string &is_deprecated(const std::string &str);
 bool is_unsafe_func(const std::string &func_name);
+bool is_compile_time_func(const std::string &func_name);
 std::string exec_system(const char *cmd);
 std::vector<std::string> resolve_binary_path(const std::string &cmd);
 std::vector<std::string> resolve_binary_path(const std::string &cmd, int pid);
@@ -127,6 +144,9 @@ void cat_file(const char *filename, size_t, std::ostream &);
 std::string str_join(const std::vector<std::string> &list,
                      const std::string &delim);
 bool is_numeric(const std::string &str);
+bool symbol_has_cpp_mangled_signature(const std::string &sym_name);
+pid_t parse_pid(const std::string &str);
+std::string hex_format_buffer(const char *buf, size_t size);
 
 // trim from end of string (right)
 inline std::string &rtrim(std::string &s)
