@@ -540,44 +540,10 @@ TEST(clang_parser, parse_fail)
 
 #ifdef HAVE_LIBBPF_BTF_DUMP
 
-#include "data/btf_data.h"
+#include "btf_common.h"
 
-class clang_parser_btf : public ::testing::Test {
- protected:
-  void SetUp() override
-  {
-    char *path = strdup("/tmp/XXXXXX");
-    if (!path)
-      return;
-
-    int fd = mkstemp(path);
-    if (fd < 0)
-    {
-      std::remove(path);
-      return;
-    }
-
-    if (write(fd, btf_data, btf_data_len) != btf_data_len)
-    {
-      close(fd);
-      std::remove(path);
-      return;
-    }
-
-    close(fd);
-    setenv("BPFTRACE_BTF", path, true);
-    path_ = path;
-  }
-
-  void TearDown() override
-  {
-    // clear the environment and remove the temp file
-    unsetenv("BPFTRACE_BTF");
-    if (path_)
-      std::remove(path_);
-  }
-
-  char *path_ = nullptr;
+class clang_parser_btf : public test_btf
+{
 };
 
 TEST_F(clang_parser_btf, btf)
@@ -709,6 +675,22 @@ TEST(clang_parser, struct_typedef)
   EXPECT_EQ(structs["max_align_t"].fields["__clang_max_align_nonce2"].offset, 16);
 }
 
+TEST(clang_parser, struct_qualifiers)
+{
+  BPFtrace bpftrace;
+  parse("struct a {int a} struct b { volatile const struct a* restrict a; "
+        "const struct a a2; };",
+        bpftrace);
+
+  StructMap &structs = bpftrace.structs_;
+  EXPECT_EQ(structs["struct b"].size, 16);
+  EXPECT_EQ(structs["struct b"].fields.size(), 2U);
+  EXPECT_EQ(structs["struct b"].fields["a"].type.type, Type::cast);
+  EXPECT_EQ(structs["struct b"].fields["a"].type.cast_type, "struct a");
+  EXPECT_EQ(structs["struct b"].fields["a2"].type.type, Type::cast);
+  EXPECT_EQ(structs["struct b"].fields["a2"].type.cast_type, "struct a");
+}
+
 } // namespace clang_parser
 } // namespace test
 } // namespace bpftrace
