@@ -2,6 +2,7 @@
 
 #include "bpffeature.h"
 #include "bpftrace.h"
+#include "child.h"
 #include "gmock/gmock.h"
 
 namespace bpftrace {
@@ -35,7 +36,11 @@ public:
   {
     (void)path;
     sym->name = name;
-    if (name[0] > 'A' && name[0] < 'z')
+    if (name == "cpp_mangled" || name == "cpp_mangled(int)")
+    {
+      return -1;
+    }
+    else if (name[0] > 'A' && name[0] < 'z')
     {
       sym->address = 12345;
       sym->size = 4;
@@ -58,8 +63,33 @@ class MockBPFfeature : public BPFfeature
 public:
   MockBPFfeature(bool has_features = true)
   {
-    has_loop_ = has_signal_ = has_get_current_cgroup_id_ =
-        has_override_return_ = has_features;
+    has_send_signal_ = std::make_optional<bool>(has_features);
+    has_get_current_cgroup_id_ = std::make_optional<bool>(has_features);
+    has_override_return_ = std::make_optional<bool>(has_features);
+    prog_kfunc_ = std::make_optional<bool>(has_features);
+    has_loop_ = std::make_optional<bool>(has_features);
+  };
+};
+
+class MockChildProc : public ChildProcBase
+{
+public:
+  MockChildProc(std::string cmd __attribute__((unused)))
+  {
+    child_pid_ = 1337;
+  };
+  ~MockChildProc(){};
+
+  void terminate(bool force __attribute__((unused)) = false) override{};
+  bool is_alive() override
+  {
+    return true;
+  };
+  void resume(void) override{};
+
+  void run(bool pause = false) override
+  {
+    (void)pause;
   };
 };
 
