@@ -10,17 +10,23 @@ namespace parser {
 
 using Printer = ast::Printer;
 
-void test_parse_failure(const std::string &input)
+void test_parse_failure(BPFtrace &bpftrace, const std::string &input)
 {
-  BPFtrace bpftrace;
   std::stringstream out;
   Driver driver(bpftrace, out);
   ASSERT_EQ(driver.parse_str(input), 1);
 }
 
-void test(const std::string &input, const std::string &output)
+void test_parse_failure(const std::string &input)
 {
   BPFtrace bpftrace;
+  test_parse_failure(bpftrace, input);
+}
+
+void test(BPFtrace &bpftrace,
+          const std::string &input,
+          const std::string &output)
+{
   Driver driver(bpftrace);
   ASSERT_EQ(driver.parse_str(input), 0);
 
@@ -30,6 +36,12 @@ void test(const std::string &input, const std::string &output)
   EXPECT_EQ(output, out.str());
 }
 
+void test(const std::string &input, const std::string &output)
+{
+  BPFtrace bpftrace;
+  test(bpftrace, input, output);
+}
+
 TEST(Parser, builtin_variables)
 {
   test("kprobe:f { pid }", "Program\n kprobe:f\n  builtin: pid\n");
@@ -45,7 +57,6 @@ TEST(Parser, builtin_variables)
   test("kprobe:f { rand }", "Program\n kprobe:f\n  builtin: rand\n");
   test("kprobe:f { ctx }", "Program\n kprobe:f\n  builtin: ctx\n");
   test("kprobe:f { comm }", "Program\n kprobe:f\n  builtin: comm\n");
-  test("kprobe:f { stack }", "Program\n kprobe:f\n  builtin: kstack\n");
   test("kprobe:f { kstack }", "Program\n kprobe:f\n  builtin: kstack\n");
   test("kprobe:f { ustack }", "Program\n kprobe:f\n  builtin: ustack\n");
   test("kprobe:f { arg0 }", "Program\n kprobe:f\n  builtin: arg0\n");
@@ -66,6 +77,83 @@ TEST(Parser, positional_param_count)
   test("kprobe:f { $# }", "Program\n kprobe:f\n  param: $#\n");
 }
 
+TEST(Parser, positional_param_attachpoint)
+{
+  BPFtrace bpftrace;
+  bpftrace.add_param("foo");
+  bpftrace.add_param("bar");
+  bpftrace.add_param("baz");
+
+  test(bpftrace,
+       "kprobe:$1 { 1 }",
+       R"PROG(Program
+ kprobe:foo
+  int: 1
+)PROG");
+
+  test(bpftrace,
+       R"PROG(kprobe:$1"here" { 1 })PROG",
+       R"PROG(Program
+ kprobe:foohere
+  int: 1
+)PROG");
+
+  test(bpftrace,
+       R"PROG(uprobe:zzzzzzz:$2 { 1 })PROG",
+       R"PROG(Program
+ uprobe:zzzzzzz:bar
+  int: 1
+)PROG");
+
+  test(bpftrace,
+       R"PROG(uprobe:$1:$2 { 1 })PROG",
+       R"PROG(Program
+ uprobe:foo:bar
+  int: 1
+)PROG");
+
+  test(bpftrace,
+       R"PROG(uprobe:$2:$1 { 1 })PROG",
+       R"PROG(Program
+ uprobe:bar:foo
+  int: 1
+)PROG");
+
+  test(bpftrace,
+       R"PROG(uprobe:"zz"$2"zz":"aa"$1 { 1 })PROG",
+       R"PROG(Program
+ uprobe:zzbarzz:aafoo
+  int: 1
+)PROG");
+
+  test(bpftrace,
+       R"PROG(uprobe:$2:"aa"$1"aa" { 1 })PROG",
+       R"PROG(Program
+ uprobe:bar:aafooaa
+  int: 1
+)PROG");
+
+  test(bpftrace,
+       R"PROG(uprobe:"$1":$2 { 1 })PROG",
+       R"PROG(Program
+ uprobe:$1:bar
+  int: 1
+)PROG");
+
+  test(bpftrace,
+       R"PROG(uprobe:aa$1:$2 { 1 })PROG",
+       R"PROG(Program
+ uprobe:aafoo:bar
+  int: 1
+)PROG");
+
+  test_parse_failure(bpftrace, R"PROG(uprobe:$1a" { 1 })PROG");
+  test_parse_failure(bpftrace, R"PROG(uprobe:$a" { 1 })PROG");
+  test_parse_failure(bpftrace, R"PROG(uprobe:$-1" { 1 })PROG");
+  test_parse_failure(bpftrace,
+                     R"PROG(uprobe:$999999999999999999999999" { 1 })PROG");
+}
+
 TEST(Parser, comment)
 {
   test("kprobe:f { /*** ***/0; }", "Program\n kprobe:f\n  int: 0\n");
@@ -648,23 +736,118 @@ TEST(Parser, if_else)
        "   variable: $s\n");
 }
 
+TEST(Parser, if_elseif)
+{
+  test("kprobe:f { if (pid > 10000) { $s = 10; } else if (pid < 10) { $s = 2; "
+       "} }",
+       "Program\n"
+       " kprobe:f\n"
+       "  if\n"
+       "   >\n"
+       "    builtin: pid\n"
+       "    int: 10000\n"
+       "   then\n"
+       "    =\n"
+       "     variable: $s\n"
+       "     int: 10\n"
+       "   else\n"
+       "    if\n"
+       "     <\n"
+       "      builtin: pid\n"
+       "      int: 10\n"
+       "     then\n"
+       "      =\n"
+       "       variable: $s\n"
+       "       int: 2\n");
+}
+
+TEST(Parser, if_elseif_else)
+{
+  test("kprobe:f { if (pid > 10000) { $s = 10; } else if (pid < 10) { $s = 2; "
+       "} else { $s = 1 } }",
+       "Program\n"
+       " kprobe:f\n"
+       "  if\n"
+       "   >\n"
+       "    builtin: pid\n"
+       "    int: 10000\n"
+       "   then\n"
+       "    =\n"
+       "     variable: $s\n"
+       "     int: 10\n"
+       "   else\n"
+       "    if\n"
+       "     <\n"
+       "      builtin: pid\n"
+       "      int: 10\n"
+       "     then\n"
+       "      =\n"
+       "       variable: $s\n"
+       "       int: 2\n"
+       "     else\n"
+       "      =\n"
+       "       variable: $s\n"
+       "       int: 1\n");
+}
+
+TEST(Parser, if_elseif_elseif_else)
+{
+  test("kprobe:f { if (pid > 10000) { $s = 10; } else if (pid < 10) { $s = 2; "
+       "} else if (pid > 999999) { $s = 0 } else { $s = 1 } }",
+       "Program\n"
+       " kprobe:f\n"
+       "  if\n"
+       "   >\n"
+       "    builtin: pid\n"
+       "    int: 10000\n"
+       "   then\n"
+       "    =\n"
+       "     variable: $s\n"
+       "     int: 10\n"
+       "   else\n"
+       "    if\n"
+       "     <\n"
+       "      builtin: pid\n"
+       "      int: 10\n"
+       "     then\n"
+       "      =\n"
+       "       variable: $s\n"
+       "       int: 2\n"
+       "     else\n"
+       "      if\n"
+       "       >\n"
+       "        builtin: pid\n"
+       "        int: 999999\n"
+       "       then\n"
+       "        =\n"
+       "         variable: $s\n"
+       "         int: 0\n"
+       "       else\n"
+       "        =\n"
+       "         variable: $s\n"
+       "         int: 1\n");
+}
+
 TEST(Parser, unroll)
 {
-  test("kprobe:sys_open { $i = 0; unroll(5) { printf(\"i: %d\\n\", $i); $i = $i + 1; } }",
+  test("kprobe:sys_open { $i = 0; unroll(5) { printf(\"i: %d\\n\", $i); $i = "
+       "$i + 1; } }",
        "Program\n"
        " kprobe:sys_open\n"
        "  =\n"
        "   variable: $i\n"
        "   int: 0\n"
-       "  unroll 5\n"
-       "   call: printf\n"
-       "    string: i: %d\\n\n"
-       "    variable: $i\n"
-       "   =\n"
-       "    variable: $i\n"
-       "    +\n"
+       "  unroll\n"
+       "   int: 5\n"
+       "   block\n"
+       "    call: printf\n"
+       "     string: i: %d\\n\n"
        "     variable: $i\n"
-       "     int: 1\n");
+       "    =\n"
+       "     variable: $i\n"
+       "     +\n"
+       "      variable: $i\n"
+       "      int: 1\n");
 }
 
 TEST(Parser, ternary_str)
@@ -774,6 +957,23 @@ TEST(Parser, uprobe)
       "Program\n"
       " uprobe:/with#hash:asdf\n"
       "  int: 1\n");
+  test("uprobe:/my/program:1234 { 1; }",
+       "Program\n"
+       " uprobe:/my/program:1234\n"
+       "  int: 1\n");
+  // Trailing alnum chars are allowed (turns the entire arg into a symbol name)
+  test("uprobe:/my/program:1234abc { 1; }",
+       "Program\n"
+       " uprobe:/my/program:1234abc\n"
+       "  int: 1\n");
+  // Test `:`s in quoted string
+  test("uprobe:/my/program:\"A::f\" { 1; }",
+       "Program\n"
+       " uprobe:/my/program:A::f\n"
+       "  int: 1\n");
+
+  test_parse_failure("uprobe:f { 1 }");
+  test_parse_failure("uprobe { 1 }");
 }
 
 TEST(Parser, usdt)
@@ -782,6 +982,15 @@ TEST(Parser, usdt)
       "Program\n"
       " usdt:/my/program:probe\n"
       "  int: 1\n");
+  // Without the escapes needed for C++ to compile:
+  //    usdt:/my/program:"\"probe\"" { 1; }
+  //
+  test("usdt:/my/program:\"\\\"probe\\\"\" { 1; }",
+       "Program\n"
+       " usdt:/my/program:\"probe\"\n"
+       "  int: 1\n");
+
+  test_parse_failure("usdt { 1 }");
 }
 
 TEST(Parser, usdt_namespaced_probe)
@@ -818,6 +1027,20 @@ TEST(Parser, begin_probe)
       "Program\n"
       " BEGIN\n"
       "  int: 1\n");
+
+  test_parse_failure("BEGIN:f { 1 }");
+  test_parse_failure("BEGIN:path:f { 1 }");
+}
+
+TEST(Parser, end_probe)
+{
+  test("END { 1 }",
+       "Program\n"
+       " END\n"
+       "  int: 1\n");
+
+  test_parse_failure("END:f { 1 }");
+  test_parse_failure("END:path:f { 1 }");
 }
 
 TEST(Parser, tracepoint_probe)
@@ -826,6 +1049,9 @@ TEST(Parser, tracepoint_probe)
       "Program\n"
       " tracepoint:sched:sched_switch\n"
       "  int: 1\n");
+
+  test_parse_failure("tracepoint:f { 1 }");
+  test_parse_failure("tracepoint { 1 }");
 }
 
 TEST(Parser, profile_probe)
@@ -834,6 +1060,11 @@ TEST(Parser, profile_probe)
       "Program\n"
       " profile:ms:997\n"
       "  int: 1\n");
+
+  test_parse_failure("profile:ms:nan { 1 }");
+  test_parse_failure("profile:f { 1 }");
+  test_parse_failure("profile { 1 }");
+  test_parse_failure("profile:s:1b { 1 }");
 }
 
 TEST(Parser, interval_probe)
@@ -842,6 +1073,8 @@ TEST(Parser, interval_probe)
       "Program\n"
       " interval:s:1\n"
       "  int: 1\n");
+
+  test_parse_failure("interval:s:1b { 1 }");
 }
 
 TEST(Parser, software_probe)
@@ -850,6 +1083,8 @@ TEST(Parser, software_probe)
       "Program\n"
       " software:faults:1000\n"
       "  int: 1\n");
+
+  test_parse_failure("software:faults:1b { 1 }");
 }
 
 TEST(Parser, hardware_probe)
@@ -858,6 +1093,20 @@ TEST(Parser, hardware_probe)
       "Program\n"
       " hardware:cache-references:1000000\n"
       "  int: 1\n");
+
+  test_parse_failure("hardware:cache-references:1b { 1 }");
+}
+
+TEST(Parser, watchpoint_probe)
+{
+  test("watchpoint::1234:8:w { 1 }",
+       "Program\n"
+       " watchpoint:1234:8:w\n"
+       "  int: 1\n");
+
+  test_parse_failure("watchpoint::1b:8:w { 1 }");
+  test_parse_failure("watchpoint::1:8a:w { 1 }");
+  test_parse_failure("watchpoint::1b:8a:w { 1 }");
 }
 
 TEST(Parser, multiple_attach_points_kprobe)
@@ -1304,6 +1553,31 @@ kprobe:f { "asdf }
   EXPECT_EQ(out.str(), expected);
 }
 
+TEST(Parser, kprobe_offset)
+{
+  test("k:fn+1 {}",
+       "Program\n"
+       " kprobe:fn+1\n");
+  test("k:fn+0x10 {}",
+       "Program\n"
+       " kprobe:fn+16\n");
+
+  test("k:\"fn.abc\"+1 {}",
+       "Program\n"
+       " kprobe:fn.abc+1\n");
+  test("k:\"fn.abc\"+0x10 {}",
+       "Program\n"
+       " kprobe:fn.abc+16\n");
+
+  test_parse_failure("k:asdf+123abc");
+}
+
+TEST(Parser, kretprobe_offset)
+{
+  // Not supported yet
+  test_parse_failure("kr:fn+1 { 1 }");
+}
+
 TEST(Parser, uprobe_offset)
 {
   test("u:./test:fn+1 {}",
@@ -1321,6 +1595,13 @@ TEST(Parser, uprobe_offset)
        " uprobe:./test:fn.abc+16\n");
 }
 
+TEST(Parser, uretprobe_offset)
+{
+  // Not supported yet
+  test_parse_failure("ur:./test:fn+1 { 1 }");
+  test_parse_failure("uretprobe:/bin/sh:f+0x10 { 1 }");
+}
+
 TEST(Parser, invalid_increment_decrement)
 {
   test_parse_failure("i:s:1 { @=5++}");
@@ -1330,6 +1611,47 @@ TEST(Parser, invalid_increment_decrement)
   test_parse_failure("i:s:1 { @=\"a\"++}");
 }
 
+TEST(Parser, long_param_overflow)
+{
+  BPFtrace bpftrace;
+  std::stringstream out;
+  Driver driver(bpftrace, out);
+  EXPECT_NO_THROW(
+      driver.parse_str("i:s:100 { @=$111111111111111111111111111 }"));
+  std::string expected = "stdin:1:11-41: ERROR: param "
+                         "$111111111111111111111111111 is out of "
+                         "integer range [1, " +
+                         std::to_string(std::numeric_limits<long>::max()) +
+                         "]\n" +
+                         "i:s:100 { @=$111111111111111111111111111 }\n" +
+                         "          ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n";
+  EXPECT_EQ(out.str(), expected);
+}
+
+TEST(Parser, empty_arguments)
+{
+  test_parse_failure("::k::vfs_open:: { 1 }");
+  test_parse_failure("k:vfs_open: { 1 }");
+  test_parse_failure(":w:0x10000000:8:rw { 1 }");
+}
+
+TEST(Parser, while_loop)
+{
+  test("i:ms:100 { $a = 0; while($a < 10) { $a++ }}",
+       R"PROG(Program
+ interval:ms:100
+  =
+   variable: $a
+   int: 0
+  while(
+   <
+    variable: $a
+    int: 10
+   )
+    variable: $a
+     ++
+)PROG");
+}
 
 } // namespace parser
 } // namespace test
