@@ -3,6 +3,7 @@
 #include "bpftrace.h"
 #include "clang_parser.h"
 #include "driver.h"
+#include "field_analyser.h"
 #include "mocks.h"
 #include "gmock/gmock.h"
 #include "gtest/gtest.h"
@@ -54,17 +55,22 @@ void test(BPFtrace &bpftrace,
           Driver &driver,
           const std::string &input,
           int expected_result = 0,
-          bool safe_mode = true)
+          bool safe_mode = true,
+          bool has_child = false)
 {
   bpftrace.safe_mode_ = safe_mode;
   ASSERT_EQ(driver.parse_str(input), 0);
 
+  ast::FieldAnalyser fields(driver.root_, bpftrace);
+  EXPECT_EQ(fields.analyse(), 0);
+
   ClangParser clang;
   clang.parse(driver.root_, bpftrace);
 
   ASSERT_EQ(driver.parse_str(input), 0);
   std::stringstream out;
-  ast::SemanticAnalyser semantics(driver.root_, bpftrace, feature, out);
+  ast::SemanticAnalyser semantics(
+      driver.root_, bpftrace, feature, out, has_child);
   std::stringstream msg;
   msg << "\nInput:\n" << input << "\n\nOutput:\n";
   EXPECT_EQ(expected_result, semantics.analyse()) << msg.str() + out.str();
@@ -76,7 +82,7 @@ void test(BPFtrace &bpftrace,
     bool safe_mode = true)
 {
   Driver driver(bpftrace);
-  MockBPFfeature feature = MockBPFfeature();
+  MockBPFfeature feature;
   test(bpftrace, feature, driver, input, expected_result, safe_mode);
 }
 
@@ -86,7 +92,7 @@ void test(Driver &driver,
     bool safe_mode = true)
 {
   auto bpftrace = get_mock_bpftrace();
-  MockBPFfeature feature = MockBPFfeature();
+  MockBPFfeature feature;
   test(*bpftrace, feature, driver, input, expected_result, safe_mode);
 }
 
@@ -101,11 +107,15 @@ void test(BPFfeature &feature,
 }
 
 void test(const std::string &input,
-    int expected_result=0,
-    bool safe_mode = true)
+          int expected_result = 0,
+          bool safe_mode = true,
+          bool has_child = false)
 {
-  MockBPFfeature feature = MockBPFfeature();
-  test(feature, input, expected_result, safe_mode);
+  MockBPFfeature feature;
+  auto bpftrace = get_mock_bpftrace();
+  Driver driver(*bpftrace);
+  test(
+      *bpftrace, feature, driver, input, expected_result, safe_mode, has_child);
 }
 
 TEST(semantic_analyser, builtin_variables)
@@ -124,7 +134,6 @@ TEST(semantic_analyser, builtin_variables)
   test("kprobe:f { rand }", 0);
   test("kprobe:f { ctx }", 0);
   test("kprobe:f { comm }", 0);
-  test("kprobe:f { stack }", 0);
   test("kprobe:f { kstack }", 0);
   test("kprobe:f { ustack }", 0);
   test("kprobe:f { arg0 }", 0);
@@ -142,14 +151,13 @@ TEST(semantic_analyser, builtin_variables)
 
 TEST(semantic_analyser, builtin_cpid)
 {
-  auto bpftrace = get_mock_bpftrace();
-  test(*bpftrace, "i:ms:100 { printf(\"%d\\n\", cpid); }", 1, false);
-  test(*bpftrace, "i:ms:100 { @=cpid }", 1, false);
-  test(*bpftrace, "i:ms:100 { $a=cpid }", 1, false);
-  bpftrace->cmd_ = "sleep 1";
-  test(*bpftrace, "i:ms:100 { printf(\"%d\\n\", cpid); }", 0, false);
-  test(*bpftrace, "i:ms:100 { @=cpid }", 0, false);
-  test(*bpftrace, "i:ms:100 { $a=cpid }", 0, false);
+  test("i:ms:100 { printf(\"%d\\n\", cpid); }", 1, false, false);
+  test("i:ms:100 { @=cpid }", 1, false, false);
+  test("i:ms:100 { $a=cpid }", 1, false, false);
+
+  test("i:ms:100 { printf(\"%d\\n\", cpid); }", 0, false, true);
+  test("i:ms:100 { @=cpid }", 0, false, true);
+  test("i:ms:100 { $a=cpid }", 0, false, true);
 }
 
 TEST(semantic_analyser, builtin_functions)
@@ -171,10 +179,10 @@ TEST(semantic_analyser, builtin_functions)
   test("kprobe:f { time() }", 0);
   test("kprobe:f { exit() }", 0);
   test("kprobe:f { str(0xffff) }", 0);
+  test("kprobe:f { buf(0xffff, 1) }", 0);
   test("kprobe:f { printf(\"hello\\n\") }", 0);
   test("kprobe:f { system(\"ls\\n\") }", 0, false /* safe_node */);
   test("kprobe:f { join(0) }", 0);
-  test("kprobe:f { sym(0xffff) }", 0);
   test("kprobe:f { ksym(0xffff) }", 0);
   test("kprobe:f { usym(0xffff) }", 0);
   test("kprobe:f { kaddr(\"sym\") }", 0);
@@ -185,6 +193,7 @@ TEST(semantic_analyser, builtin_functions)
   test("kprobe:f { ustack(1) }", 0);
   test("kprobe:f { cat(\"/proc/uptime\") }", 0);
   test("uprobe:/bin/bash:main { uaddr(\"glob_asciirange\") }", 0);
+  test("kprobe:f { cgroupid(\"/sys/fs/cgroup/unified/mycg\"); }", 0);
 }
 
 TEST(semantic_analyser, undefined_map)
@@ -216,6 +225,14 @@ TEST(semantic_analyser, consistent_map_keys)
   test("kprobe:f { @x[1,\"a\",kstack] = 0; @x[\"b\", 2, kstack]; }", 10);
 }
 
+TEST(semantic_analyser, if_statements)
+{
+  test("kprobe:f { if(1) { 123 } }", 0);
+  test("kprobe:f { if(1) { 123 } else { 456 } }", 0);
+  test("kprobe:f { if(0) { 123 } else if(1) { 456 } else { 789 } }", 0);
+  test("kprobe:f { if((int32)pid) { 123 } }", 0);
+}
+
 TEST(semantic_analyser, predicate_expressions)
 {
   test("kprobe:f / 999 / { 123 }", 0);
@@ -224,10 +241,14 @@ TEST(semantic_analyser, predicate_expressions)
   test("kprobe:f / @mymap / { @mymap = \"str\" }", 10);
 }
 
-TEST(semantic_analyser, ternary_experssions)
+TEST(semantic_analyser, ternary_expressions)
 {
   test("kprobe:f { @x = pid < 10000 ? 1 : 2 }", 0);
   test("kprobe:f { @x = pid < 10000 ? \"lo\" : \"high\" }", 0);
+  test("kprobe:f { pid < 10000 ? printf(\"lo\") : exit() }", 0);
+  test("kprobe:f { @x = pid < 10000 ? printf(\"lo\") : cat(\"/proc/uptime\") }",
+       10);
+  test("kprobe:f { pid < 10000 ? 3 : cat(\"/proc/uptime\") }", 10);
   test("kprobe:f { @x = pid < 10000 ? 1 : \"high\" }", 10);
   test("kprobe:f { @x = pid < 10000 ? \"lo\" : 2 }", 10);
 }
@@ -316,6 +337,8 @@ TEST(semantic_analyser, call_hist)
   test("kprobe:f { hist(1); }", 1);
   test("kprobe:f { $x = hist(1); }", 1);
   test("kprobe:f { @x[hist(1)] = 1; }", 1);
+  test("kprobe:f { if(hist()) { 123 } }", 1);
+  test("kprobe:f { hist() ? 0 : 1; }", 1);
 }
 
 TEST(semantic_analyser, call_lhist)
@@ -330,6 +353,8 @@ TEST(semantic_analyser, call_lhist)
   test("kprobe:f { @ = lhist(-10, -10, 10, 1); }", 10); // must be positive
   test("kprobe:f { $x = lhist(); }", 1);
   test("kprobe:f { @[lhist()] = 1; }", 1);
+  test("kprobe:f { if(lhist()) { 123 } }", 1);
+  test("kprobe:f { lhist() ? 0 : 1; }", 1);
 }
 
 TEST(semantic_analyser, call_count)
@@ -339,6 +364,8 @@ TEST(semantic_analyser, call_count)
   test("kprobe:f { count(); }", 1);
   test("kprobe:f { $x = count(); }", 1);
   test("kprobe:f { @[count()] = 1; }", 1);
+  test("kprobe:f { if(count()) { 123 } }", 1);
+  test("kprobe:f { count() ? 0 : 1; }", 1);
 }
 
 TEST(semantic_analyser, call_sum)
@@ -349,6 +376,8 @@ TEST(semantic_analyser, call_sum)
   test("kprobe:f { sum(123); }", 1);
   test("kprobe:f { $x = sum(123); }", 1);
   test("kprobe:f { @[sum(123)] = 1; }", 1);
+  test("kprobe:f { if(sum(1)) { 123 } }", 1);
+  test("kprobe:f { sum(1) ? 0 : 1; }", 1);
 }
 
 TEST(semantic_analyser, call_min)
@@ -358,6 +387,8 @@ TEST(semantic_analyser, call_min)
   test("kprobe:f { min(123); }", 1);
   test("kprobe:f { $x = min(123); }", 1);
   test("kprobe:f { @[min(123)] = 1; }", 1);
+  test("kprobe:f { if(min(1)) { 123 } }", 1);
+  test("kprobe:f { min(1) ? 0 : 1; }", 1);
 }
 
 TEST(semantic_analyser, call_max)
@@ -367,6 +398,8 @@ TEST(semantic_analyser, call_max)
   test("kprobe:f { max(123); }", 1);
   test("kprobe:f { $x = max(123); }", 1);
   test("kprobe:f { @[max(123)] = 1; }", 1);
+  test("kprobe:f { if(max(1)) { 123 } }", 1);
+  test("kprobe:f { max(1) ? 0 : 1; }", 1);
 }
 
 TEST(semantic_analyser, call_avg)
@@ -376,6 +409,8 @@ TEST(semantic_analyser, call_avg)
   test("kprobe:f { avg(123); }", 1);
   test("kprobe:f { $x = avg(123); }", 1);
   test("kprobe:f { @[avg(123)] = 1; }", 1);
+  test("kprobe:f { if(avg(1)) { 123 } }", 1);
+  test("kprobe:f { avg(1) ? 0 : 1; }", 1);
 }
 
 TEST(semantic_analyser, call_stats)
@@ -385,6 +420,8 @@ TEST(semantic_analyser, call_stats)
   test("kprobe:f { stats(123); }", 1);
   test("kprobe:f { $x = stats(123); }", 1);
   test("kprobe:f { @[stats(123)] = 1; }", 1);
+  test("kprobe:f { if(stats(1)) { 123 } }", 1);
+  test("kprobe:f { stats(1) ? 0 : 1; }", 1);
 }
 
 TEST(semantic_analyser, call_delete)
@@ -395,15 +432,20 @@ TEST(semantic_analyser, call_delete)
   test("kprobe:f { @y = delete(@x); }", 1);
   test("kprobe:f { $y = delete(@x); }", 1);
   test("kprobe:f { @[delete(@x)] = 1; }", 1);
+  test("kprobe:f { @x = 1; if(delete(@x)) { 123 } }", 10);
+  test("kprobe:f { @x = 1; delete(@x) ? 0 : 1; }", 10);
 }
 
 TEST(semantic_analyser, call_exit)
 {
   test("kprobe:f { exit(); }", 0);
   test("kprobe:f { exit(1); }", 1);
+  test("kprobe:f { @a = exit(); }", 1);
   test("kprobe:f { @a = exit(1); }", 1);
   test("kprobe:f { $a = exit(1); }", 1);
   test("kprobe:f { @[exit(1)] = 1; }", 1);
+  test("kprobe:f { if(exit()) { 123 } }", 10);
+  test("kprobe:f { exit() ? 0 : 1; }", 10);
 }
 
 TEST(semantic_analyser, call_print)
@@ -421,6 +463,27 @@ TEST(semantic_analyser, call_print)
   test("kprobe:f { @x = count(); @ = print(@x); }", 1);
   test("kprobe:f { @x = count(); $y = print(@x); }", 1);
   test("kprobe:f { @x = count(); @[print(@x)] = 1; }", 1);
+  test("kprobe:f { @x = count(); if(print(@x)) { 123 } }", 10);
+  test("kprobe:f { @x = count(); print(@x) ? 0 : 1; }", 10);
+}
+
+TEST(semantic_analyser, call_print_non_map)
+{
+  test(R"_(BEGIN { print(1) })_", 0);
+  test(R"_(BEGIN { print(comm) })_", 0);
+  test(R"_(BEGIN { print(nsecs) })_", 0);
+  test(R"_(BEGIN { print("string") })_", 0);
+  test(R"_(BEGIN { print((1, 2, "tuple")) })_", 0);
+  test(R"_(BEGIN { $x = 1; print($x) })_", 0);
+  test(R"_(BEGIN { $x = 1; $y = $x + 3; print($y) })_", 0);
+
+  test(R"_(BEGIN { print(3, 5) })_", 1);
+  test(R"_(BEGIN { print(3, 5, 2) })_", 1);
+
+  test(R"_(BEGIN { print(exit()) })_", 10);
+  test(R"_(BEGIN { print(count()) })_", 1);
+  test(R"_(BEGIN { print(ctx) })_", 10);
+  test(R"_(BEGIN { print((int8 *)0) })_", 10);
 }
 
 TEST(semantic_analyser, call_clear)
@@ -436,6 +499,8 @@ TEST(semantic_analyser, call_clear)
   test("kprobe:f { @x = count(); @ = clear(@x); }", 1);
   test("kprobe:f { @x = count(); $y = clear(@x); }", 1);
   test("kprobe:f { @x = count(); @[clear(@x)] = 1; }", 1);
+  test("kprobe:f { @x = count(); if(clear(@x)) { 123 } }", 10);
+  test("kprobe:f { @x = count(); clear(@x) ? 0 : 1; }", 10);
 }
 
 TEST(semantic_analyser, call_zero)
@@ -451,6 +516,8 @@ TEST(semantic_analyser, call_zero)
   test("kprobe:f { @x = count(); @ = zero(@x); }", 1);
   test("kprobe:f { @x = count(); $y = zero(@x); }", 1);
   test("kprobe:f { @x = count(); @[zero(@x)] = 1; }", 1);
+  test("kprobe:f { @x = count(); if(zero(@x)) { 123 } }", 10);
+  test("kprobe:f { @x = count(); zero(@x) ? 0 : 1; }", 10);
 }
 
 TEST(semantic_analyser, call_time)
@@ -463,6 +530,8 @@ TEST(semantic_analyser, call_time)
   test("kprobe:f { @[time()] = 1; }", 1);
   test("kprobe:f { time(1); }", 10);
   test("kprobe:f { $x = \"str\"; time($x); }", 10);
+  test("kprobe:f { if(time()) { 123 } }", 10);
+  test("kprobe:f { time() ? 0 : 1; }", 10);
 }
 
 TEST(semantic_analyser, call_str)
@@ -486,17 +555,36 @@ TEST(semantic_analyser, call_str_2_expr)
   test("kprobe:f { @x = str(arg0, arg1); }", 0);
 }
 
-TEST(semantic_analyser, call_sym)
+TEST(semantic_analyser, call_buf)
+{
+  test("kprobe:f { buf(arg0, 1); }", 0);
+  test("kprobe:f { @x = buf(arg0, 1); }", 0);
+  test("kprobe:f { $x = buf(arg0, 1); }", 0);
+  test("kprobe:f { buf(); }", 1);
+  test("kprobe:f { buf(\"hello\"); }", 1);
+  test("struct x { int c[4] }; kprobe:f { $foo = (struct x*)0; @x = "
+       "buf($foo->c); }",
+       0);
+}
+
+TEST(semantic_analyser, call_buf_lit)
+{
+  test("kprobe:f { @x = buf(arg0, 3); }", 0);
+  test("kprobe:f { buf(arg0, \"hello\"); }", 10);
+}
+
+TEST(semantic_analyser, call_buf_expr)
+{
+  test("kprobe:f { buf(arg0, arg1); }", 0);
+  test("kprobe:f { @x = buf(arg0, arg1); }", 0);
+}
+
+TEST(semantic_analyser, call_ksym)
 {
   test("kprobe:f { ksym(arg0); }", 0);
   test("kprobe:f { @x = ksym(arg0); }", 0);
   test("kprobe:f { ksym(); }", 1);
   test("kprobe:f { ksym(\"hello\"); }", 1);
-
-  test("kprobe:f { sym(arg0); }", 0);
-  test("kprobe:f { @x = sym(arg0); }", 0);
-  test("kprobe:f { sym(); }", 1);
-  test("kprobe:f { sym(\"hello\"); }", 1);
 }
 
 TEST(semantic_analyser, call_usym)
@@ -525,6 +613,9 @@ TEST(semantic_analyser, call_ntop)
   test(structs + "kprobe:f { @x = ntop(((struct inet*)0)->ipv4); }", 0);
   test(structs + "kprobe:f { @x = ntop(((struct inet*)0)->ipv6); }", 0);
 
+  // Regression test that ntop can use arguments from the prog context
+  test("tracepoint:tcp:some_tcp_tp { ntop(args->saddr_v6); }", 0);
+
   test("kprobe:f { ntop(); }", 1);
   test("kprobe:f { ntop(2, \"hello\"); }", 1);
   test("kprobe:f { ntop(\"hello\"); }", 1);
@@ -575,12 +666,22 @@ TEST(semantic_analyser, call_uaddr)
   {
     auto v = static_cast<ast::AssignVarStatement *>(
         driver.root_->probes->at(0)->stmts->at(i));
-    EXPECT_EQ(SizedType(Type::integer, 8, false), v->var->type);
+    EXPECT_EQ(CreateUInt64(), v->var->type);
     EXPECT_EQ(true, v->var->type.is_pointer);
     EXPECT_EQ((unsigned long int)sizes.at(i), v->var->type.pointee_size);
   }
 }
 
+TEST(semantic_analyser, call_cgroupid)
+{
+  // Handle args above STRING_SIZE
+  test("kprobe:f { cgroupid("
+       //          1         2         3         4         5         6
+       "\"123456789/123456789/123456789/123456789/123456789/123456789/12345\""
+       "); }",
+       0);
+}
+
 TEST(semantic_analyser, call_reg)
 {
   test("kprobe:f { reg(\"ip\"); }", 0);
@@ -613,6 +714,8 @@ TEST(semantic_analyser, call_cat)
   test("kprobe:f { @x = cat(\"/proc/loadavg\"); }", 1);
   test("kprobe:f { $x = cat(\"/proc/loadavg\"); }", 1);
   test("kprobe:f { @[cat(\"/proc/loadavg\")] = 1; }", 1);
+  test("kprobe:f { if(cat(\"/proc/loadavg\")) { 123 } }", 10);
+  test("kprobe:f { cat(\"/proc/loadavg\") ? 0 : 1; }", 10);
 }
 
 TEST(semantic_analyser, call_stack)
@@ -710,7 +813,7 @@ TEST(semantic_analyser, array_access) {
        0);
   auto assignment = static_cast<ast::AssignMapStatement *>(
       driver.root_->probes->at(0)->stmts->at(1));
-  EXPECT_EQ(SizedType(Type::integer, 8, true), assignment->map->type);
+  EXPECT_EQ(CreateInt64(), assignment->map->type);
 }
 
 TEST(semantic_analyser, variable_type)
@@ -718,7 +821,7 @@ TEST(semantic_analyser, variable_type)
   BPFtrace bpftrace;
   Driver driver(bpftrace);
   test(driver, "kprobe:f { $x = 1 }", 0);
-  SizedType st(Type::integer, 8, true);
+  auto st = CreateInt64();
   auto assignment = static_cast<ast::AssignVarStatement*>(driver.root_->probes->at(0)->stmts->at(0));
   EXPECT_EQ(st, assignment->var->type);
 }
@@ -726,8 +829,19 @@ TEST(semantic_analyser, variable_type)
 TEST(semantic_analyser, unroll)
 {
   test("kprobe:f { $i = 0; unroll(5) { printf(\"i: %d\\n\", $i); $i = $i + 1; } }", 0);
-  test("kprobe:f { $i = 0; unroll(21) { printf(\"i: %d\\n\", $i); $i = $i + 1; } }", 1);
+  test("kprobe:f { $i = 0; unroll(101) { printf(\"i: %d\\n\", $i); $i = $i + "
+       "1; } }",
+       1);
   test("kprobe:f { $i = 0; unroll(0) { printf(\"i: %d\\n\", $i); $i = $i + 1; } }", 1);
+
+  BPFtrace bpftrace;
+  bpftrace.add_param("10");
+  bpftrace.add_param("hello");
+  bpftrace.add_param("101");
+  test(bpftrace, "kprobe:f { unroll($#) { printf(\"hi\\n\"); } }", 0);
+  test(bpftrace, "kprobe:f { unroll($1) { printf(\"hi\\n\"); } }", 0);
+  test(bpftrace, "kprobe:f { unroll($2) { printf(\"hi\\n\"); } }", 1);
+  test(bpftrace, "kprobe:f { unroll($3) { printf(\"hi\\n\"); } }", 1);
 }
 
 TEST(semantic_analyser, map_integer_sizes)
@@ -739,8 +853,8 @@ TEST(semantic_analyser, map_integer_sizes)
 
   auto var_assignment = static_cast<ast::AssignVarStatement*>(driver.root_->probes->at(0)->stmts->at(0));
   auto map_assignment = static_cast<ast::AssignMapStatement*>(driver.root_->probes->at(0)->stmts->at(1));
-  EXPECT_EQ(SizedType(Type::integer, 4, true), var_assignment->var->type);
-  EXPECT_EQ(SizedType(Type::integer, 8, true), map_assignment->map->type);
+  EXPECT_EQ(CreateInt32(), var_assignment->var->type);
+  EXPECT_EQ(CreateInt64(), map_assignment->map->type);
 }
 
 TEST(semantic_analyser, unop_dereference)
@@ -788,6 +902,9 @@ TEST(semantic_analyser, printf)
   test("kprobe:f { printf(\"%A\", comm) }", 10);
   test("kprobe:f { @x = printf(\"hi\") }", 1);
   test("kprobe:f { $x = printf(\"hi\") }", 1);
+  test("kprobe:f { printf(\"%d %d %d %d %d %d %d %d %d\", 1, 2, 3, 4, 5, 6, 7, "
+       "8, 9); }",
+       0);
 }
 
 TEST(semantic_analyser, system)
@@ -879,6 +996,17 @@ TEST(semantic_analyser, printf_bad_format_string)
   test("kprobe:f { printf(\"%s\", arg0) }", 10);
 }
 
+TEST(semantic_analyser, printf_format_buf)
+{
+  test("kprobe:f { printf(\"%r\", buf(\"mystr\", 5)) }", 0);
+}
+
+TEST(semantic_analyser, printf_bad_format_buf)
+{
+  test("kprobe:f { printf(\"%r\", \"mystr\") }", 10);
+  test("kprobe:f { printf(\"%r\", arg0) }", 10);
+}
+
 TEST(semantic_analyser, printf_format_multi)
 {
   test("kprobe:f { printf(\"%d %d %s\", 1, 2, \"mystr\") }", 0);
@@ -908,12 +1036,7 @@ TEST(semantic_analyser, join_delimiter)
 TEST(semantic_analyser, kprobe)
 {
   test("kprobe:f { 1 }", 0);
-  test("kprobe:path:f { 1 }", 1);
-  test("kprobe { 1 }", 1);
-
   test("kretprobe:f { 1 }", 0);
-  test("kretprobe:path:f { 1 }", 1);
-  test("kretprobe { 1 }", 1);
 }
 
 TEST(semantic_analyser, uprobe)
@@ -927,8 +1050,6 @@ TEST(semantic_analyser, uprobe)
   test("uprobe:sh:f { 1 }", 0);
   test("uprobe:/notexistfile:f { 1 }", 1);
   test("uprobe:notexistfile:f { 1 }", 1);
-  test("uprobe:f { 1 }", 1);
-  test("uprobe { 1 }", 1);
 
   test("uretprobe:/bin/sh:f { 1 }", 0);
   test("ur:/bin/sh:f { 1 }", 0);
@@ -936,12 +1057,8 @@ TEST(semantic_analyser, uprobe)
   test("ur:sh:f { 1 }", 0);
   test("uretprobe:/bin/sh:0x10 { 1 }", 0);
   test("ur:/bin/sh:0x10 { 1 }", 0);
-  test("uretprobe:/bin/sh:f+0x10 { 1 }", 1);
-  test("ur:/bin/sh:f+0x10 { 1 }", 1);
   test("uretprobe:/notexistfile:f { 1 }", 1);
   test("uretprobe:notexistfile:f { 1 }", 1);
-  test("uretprobe:f { 1 }", 1);
-  test("uretprobe { 1 }", 1);
 }
 
 TEST(semantic_analyser, usdt)
@@ -951,27 +1068,20 @@ TEST(semantic_analyser, usdt)
   test("usdt:/bin/sh:namespace:probe { 1 }", 0);
   test("usdt:/notexistfile:probe { 1 }", 1);
   test("usdt:notexistfile:probe { 1 }", 1);
-  test("usdt { 1 }", 1);
 }
 
 TEST(semantic_analyser, begin_end_probes)
 {
   test("BEGIN { 1 }", 0);
-  test("BEGIN:f { 1 }", 1);
-  test("BEGIN:path:f { 1 }", 1);
   test("BEGIN { 1 } BEGIN { 2 }", 10);
 
   test("END { 1 }", 0);
-  test("END:f { 1 }", 1);
-  test("END:path:f { 1 }", 1);
   test("END { 1 } END { 2 }", 10);
 }
 
 TEST(semantic_analyser, tracepoint)
 {
   test("tracepoint:category:event { 1 }", 0);
-  test("tracepoint:f { 1 }", 1);
-  test("tracepoint { 1 }", 1);
 }
 
 TEST(semantic_analyser, watchpoint)
@@ -993,9 +1103,9 @@ TEST(semantic_analyser, args_builtin_wrong_use)
   test("END { args->foo }", 1);
   test("kprobe:f { args->foo }", 1);
   test("kretprobe:f { args->foo }", 1);
-  test("uprobe:f { args->foo }", 1);
-  test("uretprobe:f { args->foo }", 1);
-  test("profile:f { args->foo }", 1);
+  test("uprobe:/bin/sh:f { args->foo }", 1);
+  test("uretprobe:/bin/sh/:f { args->foo }", 1);
+  test("profile:ms:1 { args->foo }", 1);
   test("usdt:sh:probe { args->foo }", 1);
   test("profile:ms:100 { args->foo }", 1);
   test("hardware:cache-references:1000000 { args->foo }", 1);
@@ -1009,10 +1119,16 @@ TEST(semantic_analyser, profile)
   test("profile:s:10 { 1 }", 0);
   test("profile:ms:100 { 1 }", 0);
   test("profile:us:100 { 1 }", 0);
-  test("profile:ms:nan { 1 }", 1);
   test("profile:unit:100 { 1 }", 1);
-  test("profile:f { 1 }", 1);
-  test("profile { 1 }", 1);
+}
+
+TEST(semantic_analyser, interval)
+{
+  test("interval:hz:997 { 1 }", 0);
+  test("interval:s:10 { 1 }", 0);
+  test("interval:ms:100 { 1 }", 0);
+  test("interval:us:100 { 1 }", 0);
+  test("interval:unit:100 { 1 }", 1);
 }
 
 TEST(semantic_analyser, variable_cast_types)
@@ -1140,6 +1256,7 @@ TEST(semantic_analyser, positional_parameters)
   BPFtrace bpftrace;
   bpftrace.add_param("123");
   bpftrace.add_param("hello");
+  bpftrace.add_param("0x123");
 
   test(bpftrace, "kprobe:f { printf(\"%d\", $0); }", 1);
   test(bpftrace, "kprobe:f { printf(\"%s\", str($0)); }", 1);
@@ -1150,12 +1267,25 @@ TEST(semantic_analyser, positional_parameters)
   test(bpftrace, "kprobe:f { printf(\"%s\", str($2)); }", 0);
   test(bpftrace, "kprobe:f { printf(\"%d\", $2); }", 10);
 
-  // Parameters are not required to exist to be used:
-  test(bpftrace, "kprobe:f { printf(\"%s\", str($3)); }", 0);
   test(bpftrace, "kprobe:f { printf(\"%d\", $3); }", 0);
 
+  // Parameters are not required to exist to be used:
+  test(bpftrace, "kprobe:f { printf(\"%s\", str($4)); }", 0);
+  test(bpftrace, "kprobe:f { printf(\"%d\", $4); }", 0);
+
   test(bpftrace, "kprobe:f { printf(\"%d\", $#); }", 0);
   test(bpftrace, "kprobe:f { printf(\"%s\", str($#)); }", 10);
+
+  Driver driver(bpftrace);
+  test(driver, "k:f { $1 }", 0);
+  auto stmt = static_cast<ast::ExprStatement *>(
+      driver.root_->probes->at(0)->stmts->at(0));
+  auto pp = static_cast<ast::PositionalParameter *>(stmt->expr);
+  EXPECT_EQ(CreateInt64(), pp->type);
+  EXPECT_TRUE(pp->is_literal);
+
+  bpftrace.add_param("0999");
+  test(bpftrace, "kprobe:f { printf(\"%d\", $4); }", 10);
 }
 
 TEST(semantic_analyser, macros)
@@ -1257,10 +1387,10 @@ TEST(semantic_analyser, cast_sign)
   auto us = static_cast<ast::AssignVarStatement*>(driver.root_->probes->at(0)->stmts->at(2));
   auto l  = static_cast<ast::AssignVarStatement*>(driver.root_->probes->at(0)->stmts->at(3));
   auto ul = static_cast<ast::AssignVarStatement*>(driver.root_->probes->at(0)->stmts->at(4));
-  EXPECT_EQ(SizedType(Type::integer, 4, true),  s->var->type);
-  EXPECT_EQ(SizedType(Type::integer, 4, false), us->var->type);
-  EXPECT_EQ(SizedType(Type::integer, 8, true),  l->var->type);
-  EXPECT_EQ(SizedType(Type::integer, 8, false), ul->var->type);
+  EXPECT_EQ(CreateInt32(), s->var->type);
+  EXPECT_EQ(CreateUInt32(), us->var->type);
+  EXPECT_EQ(CreateInt64(), l->var->type);
+  EXPECT_EQ(CreateUInt64(), ul->var->type);
 }
 
 TEST(semantic_analyser, binop_sign)
@@ -1284,11 +1414,11 @@ TEST(semantic_analyser, binop_sign)
 
     test(driver, prog, 0);
     auto varA = static_cast<ast::AssignVarStatement*>(driver.root_->probes->at(0)->stmts->at(1));
-    EXPECT_EQ(SizedType(Type::integer, 8, true), varA->var->type);
+    EXPECT_EQ(CreateInt64(), varA->var->type);
     auto varB = static_cast<ast::AssignVarStatement*>(driver.root_->probes->at(0)->stmts->at(2));
-    EXPECT_EQ(SizedType(Type::integer, 8, false), varB->var->type);
+    EXPECT_EQ(CreateUInt64(), varB->var->type);
     auto varC = static_cast<ast::AssignVarStatement*>(driver.root_->probes->at(0)->stmts->at(3));
-    EXPECT_EQ(SizedType(Type::integer, 8, false), varC->var->type);
+    EXPECT_EQ(CreateUInt64(), varC->var->type);
   }
 }
 
@@ -1405,7 +1535,7 @@ TEST(semantic_analyser, override)
 
   // Probe types
   test("kr:f { override(-1); }", 1, false);
-  test("u:f { override(-1); }", 1, false);
+  test("u:/bin/sh:f { override(-1); }", 1, false);
   test("t:syscalls:sys_enter_openat { override(-1); }", 1, false);
   test("i:s:1 { override(-1); }", 1, false);
   test("p:hz:1 { override(-1); }", 1, false);
@@ -1425,6 +1555,44 @@ TEST(semantic_analyser, struct_member_keywords)
   }
 }
 
+TEST(semantic_analyser, jumps)
+{
+  test("i:s:1 { return; }", 0);
+  // must be used in loops
+  test("i:s:1 { break; }", 1);
+  test("i:s:1 { continue; }", 1);
+}
+
+TEST(semantic_analyser, while_loop)
+{
+  test("i:s:1 { $a = 1; while ($a < 10) { $a++ }}", 0);
+  test("i:s:1 { $a = 1; while (1) { if($a > 50) { break } $a++ }}", 0);
+  test("i:s:1 { $a = 1; while ($a < 10) { $a++ }}", 0);
+  test("i:s:1 { $a = 1; while (1) { if($a > 50) { break } $a++ }}", 0);
+  test("i:s:1 { $a = 1; while (1) { if($a > 50) { return } $a++ }}", 0);
+  test(R"PROG(
+i:s:1 {
+  $a = 1;
+  while ($a < 10) {
+    $a++; $j=0;
+    while ($j < 10) {
+      $j++;
+    }
+  }
+})PROG",
+       0);
+
+  test_for_warning("i:s:1 { $a = 1; while ($a < 10) { break; $a++ }}",
+                   "code after a 'break'");
+  test_for_warning("i:s:1 { $a = 1; while ($a < 10) { continue; $a++ }}",
+                   "code after a 'continue'");
+  test_for_warning("i:s:1 { $a = 1; while ($a < 10) { return; $a++ }}",
+                   "code after a 'return'");
+
+  test_for_warning("i:s:1 { $a = 1; while ($a < 10) { @=$a++; print(@); }}",
+                   "'print()' in a loop");
+}
+
 TEST(semantic_analyser, builtin_args)
 {
   auto bpftrace = get_mock_bpftrace();
@@ -1436,6 +1604,220 @@ TEST(semantic_analyser, builtin_args)
   test(*bpftrace, "t:sched:sched_one { args->not_a_field }", 1);
 }
 
+TEST(semantic_analyser, type_ctx)
+{
+  BPFtrace bpftrace;
+  Driver driver(bpftrace);
+  std::string structs = "struct c {char c} struct x { long a; short b[4]; "
+                        "struct c c; struct c *d;}";
+  test(driver,
+       structs + "kprobe:f { $x = (struct x*)ctx; $a = $x->a; $b = $x->b[0]; "
+                 "$c = $x->c.c; $d = $x->d->c;}",
+       0);
+  auto &stmts = driver.root_->probes->at(0)->stmts;
+
+  // $x = (struct x*)ctx;
+  auto assignment = static_cast<ast::AssignVarStatement *>(stmts->at(0));
+  EXPECT_EQ(SizedType(Type::ctx, 8, false), assignment->var->type);
+
+  // $a = $x->a;
+  assignment = static_cast<ast::AssignVarStatement *>(stmts->at(1));
+  EXPECT_EQ(CreateInt64(), assignment->var->type);
+  auto fieldaccess = static_cast<ast::FieldAccess *>(assignment->expr);
+  EXPECT_EQ(CreateInt64(), fieldaccess->type);
+  auto unop = static_cast<ast::Unop *>(fieldaccess->expr);
+  EXPECT_EQ(SizedType(Type::ctx, 32, false), unop->type);
+  auto var = static_cast<ast::Variable *>(unop->expr);
+  EXPECT_EQ(SizedType(Type::ctx, 8, false), var->type);
+
+  // $b = $x->b[0];
+  assignment = static_cast<ast::AssignVarStatement *>(stmts->at(2));
+  EXPECT_EQ(CreateInt16(), assignment->var->type);
+  auto arrayaccess = static_cast<ast::ArrayAccess *>(assignment->expr);
+  EXPECT_EQ(CreateInt16(), arrayaccess->type);
+  fieldaccess = static_cast<ast::FieldAccess *>(arrayaccess->expr);
+  EXPECT_EQ(SizedType(Type::ctx, 4, false), fieldaccess->type);
+  unop = static_cast<ast::Unop *>(fieldaccess->expr);
+  EXPECT_EQ(SizedType(Type::ctx, 32, false), unop->type);
+  var = static_cast<ast::Variable *>(unop->expr);
+  EXPECT_EQ(SizedType(Type::ctx, 8, false), var->type);
+
+  // $c = $x->c.c;
+  assignment = static_cast<ast::AssignVarStatement *>(stmts->at(3));
+  EXPECT_EQ(CreateInt8(), assignment->var->type);
+  fieldaccess = static_cast<ast::FieldAccess *>(assignment->expr);
+  EXPECT_EQ(CreateInt8(), fieldaccess->type);
+  fieldaccess = static_cast<ast::FieldAccess *>(fieldaccess->expr);
+  EXPECT_EQ(SizedType(Type::ctx, 1, false), fieldaccess->type);
+  unop = static_cast<ast::Unop *>(fieldaccess->expr);
+  EXPECT_EQ(SizedType(Type::ctx, 32, false), unop->type);
+  var = static_cast<ast::Variable *>(unop->expr);
+  EXPECT_EQ(SizedType(Type::ctx, 8, false), var->type);
+
+  // $d = $x->d->c;
+  assignment = static_cast<ast::AssignVarStatement *>(stmts->at(4));
+  EXPECT_EQ(CreateInt8(), assignment->var->type);
+  fieldaccess = static_cast<ast::FieldAccess *>(assignment->expr);
+  EXPECT_EQ(CreateInt8(), fieldaccess->type);
+  unop = static_cast<ast::Unop *>(fieldaccess->expr);
+  EXPECT_EQ(CreateCast(8), unop->type);
+  fieldaccess = static_cast<ast::FieldAccess *>(unop->expr);
+  EXPECT_EQ(CreateCast(64), fieldaccess->type);
+  unop = static_cast<ast::Unop *>(fieldaccess->expr);
+  EXPECT_EQ(SizedType(Type::ctx, 32, false), unop->type);
+  var = static_cast<ast::Variable *>(unop->expr);
+  EXPECT_EQ(SizedType(Type::ctx, 8, false), var->type);
+
+  test(driver, "k:f, kr:f { @ = (uint64)ctx; }", 0);
+  test(driver, "k:f, i:s:1 { @ = (uint64)ctx; }", 1);
+  test(driver, "t:sched:sched_one { @ = (uint64)ctx; }", 1);
+}
+
+// Basic functionality test
+TEST(semantic_analyser, tuple)
+{
+  test(R"_(BEGIN { $t = (1)})_", 0);
+  test(R"_(BEGIN { $t = (1, 2); $v = $t;})_", 0);
+  test(R"_(BEGIN { $t = (1, 2, "string")})_", 0);
+  test(R"_(BEGIN { $t = (1, 2, "string"); $t = (3, 4, "other"); })_", 0);
+  test(R"_(BEGIN { $t = (1, kstack()) })_", 0);
+  test(R"_(BEGIN { $t = (1, (2,3)) })_", 0);
+
+  test(R"_(BEGIN { @t = (1)})_", 0);
+  test(R"_(BEGIN { @t = (1, 2); @v = @t;})_", 0);
+  test(R"_(BEGIN { @t = (1, 2, "string")})_", 0);
+  test(R"_(BEGIN { @t = (1, 2, "string"); @t = (3, 4, "other"); })_", 0);
+  test(R"_(BEGIN { @t = (1, kstack()) })_", 0);
+  test(R"_(BEGIN { @t = (1, (2,3)) })_", 0);
+
+  test(R"_(BEGIN { $t = (1, 2); $t = (4, "other"); })_", 10);
+  test(R"_(BEGIN { $t = (1, 2); $t = 5; })_", 1);
+  test(R"_(BEGIN { $t = (1, count()) })_", 1);
+
+  test(R"_(BEGIN { @t = (1, 2); @t = (4, "other"); })_", 10);
+  test(R"_(BEGIN { @t = (1, 2); @t = 5; })_", 1);
+  test(R"_(BEGIN { @t = (1, count()) })_", 1);
+}
+
+TEST(semantic_analyser, tuple_indexing)
+{
+  test(R"_(BEGIN { (1,2).0 })_", 0);
+  test(R"_(BEGIN { (1,2).1 })_", 0);
+  test(R"_(BEGIN { (1,2,3).2 })_", 0);
+  test(R"_(BEGIN { $t = (1,2,3).0 })_", 0);
+  test(R"_(BEGIN { $t = (1,2,3); $v = $t.0; })_", 0);
+
+  test(R"_(BEGIN { (1,2,3).3 })_", 10);
+  test(R"_(BEGIN { (1,2,3).9999999999999 })_", 10);
+}
+
+// More in depth inspection of AST
+TEST(semantic_analyser, tuple_assign_var)
+{
+  BPFtrace bpftrace;
+  Driver driver(bpftrace);
+  test(driver, R"_(BEGIN { $t = (1, "str"); $t = (4, "other"); })_", 0);
+
+  auto &stmts = driver.root_->probes->at(0)->stmts;
+
+  // $t = (1, "str");
+  auto assignment = static_cast<ast::AssignVarStatement *>(stmts->at(0));
+  auto ty = SizedType(Type::tuple, 8 + 64, false);
+  ty.tuple_elems.emplace_back(Type::integer, 8, true);
+  ty.tuple_elems.emplace_back(Type::string, 64, false);
+  EXPECT_EQ(ty, assignment->var->type);
+
+  // $t = (4, "other");
+  assignment = static_cast<ast::AssignVarStatement *>(stmts->at(1));
+  ty = SizedType(Type::tuple, 8 + 64, false);
+  ty.tuple_elems.emplace_back(Type::integer, 8, true);
+  ty.tuple_elems.emplace_back(Type::string, 64, false);
+  EXPECT_EQ(ty, assignment->var->type);
+}
+
+// More in depth inspection of AST
+TEST(semantic_analyser, tuple_assign_map)
+{
+  BPFtrace bpftrace;
+  Driver driver(bpftrace);
+  test(driver, R"_(BEGIN { @ = (1, 3, 3, 7); @ = (0, 0, 0, 0); })_", 0);
+
+  auto &stmts = driver.root_->probes->at(0)->stmts;
+
+  // $t = (1, 3, 3, 7);
+  auto assignment = static_cast<ast::AssignMapStatement *>(stmts->at(0));
+  auto ty = SizedType(Type::tuple, 4 * 8, false);
+  ty.tuple_elems.emplace_back(Type::integer, 8, true);
+  ty.tuple_elems.emplace_back(Type::integer, 8, false);
+  ty.tuple_elems.emplace_back(Type::integer, 8, false);
+  ty.tuple_elems.emplace_back(Type::integer, 8, false);
+  EXPECT_EQ(ty, assignment->map->type);
+
+  // $t = (0, 0, 0, 0);
+  assignment = static_cast<ast::AssignMapStatement *>(stmts->at(1));
+  ty = SizedType(Type::tuple, 4 * 8, false);
+  ty.tuple_elems.emplace_back(Type::integer, 8, true);
+  ty.tuple_elems.emplace_back(Type::integer, 8, true);
+  ty.tuple_elems.emplace_back(Type::integer, 8, true);
+  ty.tuple_elems.emplace_back(Type::integer, 8, true);
+  EXPECT_EQ(ty, assignment->map->type);
+}
+
+// More in depth inspection of AST
+TEST(semantic_analyser, tuple_nested)
+{
+  BPFtrace bpftrace;
+  Driver driver(bpftrace);
+  test(driver, R"_(BEGIN { $t = (1,(1,2)); })_", 0);
+
+  auto &stmts = driver.root_->probes->at(0)->stmts;
+
+  // $t = (1, "str");
+  auto assignment = static_cast<ast::AssignVarStatement *>(stmts->at(0));
+  auto ty = SizedType(Type::tuple, 3 * 8, false);
+  ty.tuple_elems.emplace_back(Type::integer, 8, true);
+  ty.tuple_elems.emplace_back(Type::tuple, 2 * 8, false);
+  auto ty_inner = SizedType(Type::tuple, 2 * 8, false);
+  ty_inner.tuple_elems.emplace_back(Type::tuple, 2 * 8, false);
+  ty.tuple_elems.emplace_back(std::move(ty_inner));
+  EXPECT_EQ(ty, assignment->var->type);
+}
+
+TEST(semantic_analyser, multi_pass_type_inference_zero_size_int)
+{
+  auto bpftrace = get_mock_bpftrace();
+  // The first pass on processing the Unop does not have enough information
+  // to figure out size of `@i` yet. The analyzer figures out the size
+  // after seeing the `@i++`. On the second pass the correct size is
+  // determined.
+  test(*bpftrace, "BEGIN { if (!@i) { @i++; } }", 0);
+}
+
+#ifdef HAVE_LIBBPF_BTF_DUMP
+
+#include "btf_common.h"
+
+class semantic_analyser_btf : public test_btf
+{
+};
+
+TEST_F(semantic_analyser_btf, kfunc)
+{
+  test("kfunc:func_1 { 1 }", 0);
+  test("kretfunc:func_1 { 1 }", 0);
+  test("kfunc:func_1 { $x = args->a; $y = args->foo1; }", 0);
+  test("kretfunc:func_1 { $x = retval; }", 0);
+  test("kretfunc:func_1 { $x = args->foo; }", 1);
+}
+
+TEST_F(semantic_analyser_btf, short_name)
+{
+  test("f:func_1 { 1 }", 0);
+  test("fr:func_1 { 1 }", 0);
+}
+
+#endif // HAVE_LIBBPF_BTF_DUMP
+
 } // namespace semantic_analyser
 } // namespace test
 } // namespace bpftrace
